/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.curves;

import java.util.ArrayList;
import java.util.List;

import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;

public class ArrivalCurve extends Curve {
	public static final ArrivalCurve NULL_ARRIVAL_CURVE = new ArrivalCurve( Curve.NULL_CURVE );
	
	private List<ArrivalCurve> token_buckets;
	
	@SuppressWarnings("unused")
	private ArrivalCurve() {
		token_buckets = null;
	}
	
	public ArrivalCurve( int segment_count ) {
		super( segment_count );
		token_buckets = null;
	}
		
	public ArrivalCurve( Curve curve ) {
		super();

		if( Configuration.performChecks() && !this.isWideSenseIncreasing() ) {
			throw new RuntimeException("Arrival curves can only be created from wide sense increasing curves.");
		}

		token_buckets = null;
		alias = curve.getAlias();
		label = curve.getLabel();

		int i = 0;
		if ( curve.getSegment(0).y0.greater( Num.ZERO ) ) {
			segments = new LinearSegment[curve.getSegmentCount()+1];
			
			segments[0] = LinearSegment.createNullSegment();
			i++;
			
			segments[1] = curve.getSegment( 0 ).copy();
			segments[1].leftopen = true;
			i++;
		} else {
			segments = new LinearSegment[curve.getSegmentCount()];
		}
		
		for ( ; i < curve.getSegmentCount(); i++ ) {
			segments[i] = curve.getSegment( i ).copy();
		}
	}
	
	/**
	 * Creates a null arrival curve.
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createNullArrival() {
		return NULL_ARRIVAL_CURVE.copy();
	}

	/**
	 * Creates a new token bucket curve.
	 * @param r the rate
	 * @param b the burstiness
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createTokenBucket( Num r, Num b ) {
		return new ArrivalCurve( Curve.createTokenBucket(r, b) );
	}

	/**
	 * Creates a new token bucket curve.
	 * @param r the rate
	 * @param b the burstiness
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createTokenBucket( double r, double b ) {
		return new ArrivalCurve( Curve.createTokenBucket(r, b) );
	}

	/**
	 * Creates a new curve from a list of token bucket curves.
	 * @param token_buckets a list of token bucket curves
	 * @return a <code>Curve</code> instance
	 */
	public static ArrivalCurve createFromTokenBuckets( List<Curve> token_buckets ) {
		return new ArrivalCurve( Curve.createFromTokenBuckets( token_buckets ) );
	}
	
	public Num getBurst() {
		return fLimitRight( Num.ZERO );
	}
	
	public ArrivalCurve copy() {
		return new ArrivalCurve( this );
	}

	public static ArrivalCurve add( ArrivalCurve c1, ArrivalCurve c2 ) {
		return new ArrivalCurve( Curve.add( c1, c2 ) );
	}

	public static ArrivalCurve add( ArrivalCurve c1, Num dy ) {
		return new ArrivalCurve( Curve.add( c1, dy ) );
	}
	
	/**
	 * Returns a copy of this curve that is shifted to the right by <code>dx</code>,
	 * i.e. g(x) = f(x-dx).
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public ArrivalCurve shiftRight( Num dx ) {
		return new ArrivalCurve( Curve.shiftRight( this, dx ) );
	}
	
	/**
	 * Returns a copy of this curve that is shifted to the left by <code>dx</code>,
	 * i.e. g(x) = f(x+dx). Note that the new curve is clipped at the y-axis so
	 * that in most cases <code>c.shiftLeftClipping(dx).shiftRight(dx) != c</code>!
	 * @param dx the offset to shift the curve.
	 * @return the shifted curve.
	 */
	public ArrivalCurve shiftLeftClipping( Num dx ) {
		return new ArrivalCurve( Curve.shiftLeftClipping( this, dx ) );
	}

	/**
	 * Decomposes this curve into a list of token bucket curves and
	 * stores this list in the curve's <code>token_buckets</code> field.<br/>
	 * Note: Curve must be concave.
	 */
	private void decomposeIntoTokenBuckets() {
		if( Configuration.performChecks() && !this.isConcave() ) {
			throw new RuntimeException("Can only decompose concave arrival curves into token buckets.");
		}

		token_buckets = new ArrayList<ArrivalCurve>();
		for (int i = 0; i < segments.length; i++) {
			if (isDiscontinuity(i)) {
				continue;
			}
			Num r = segments[i].grad;
			Num b = Num.sub( segments[i].y0, Num.mult( segments[i].x0, segments[i].grad ) );
			token_buckets.add( createTokenBucket( r, b ) );
		}
	}

	/**
	 * Returns the number of token buckets the curve can be decomposed into.
	 * @return the number of token buckets
	 */
	public int getTBComponentCount() {
		if ( token_buckets == null ) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets.size();
	}

	/**
	 * Returns a list of token bucket curves that this curve can be
	 * decomposed into.
	 * @return the list of token buckets
	 */
	public final List<ArrivalCurve> getTBComponents() {
		if ( token_buckets == null ) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets;
	}

	/**
	 * Returns the <code>i</code>th token bucket curve that this curve can be
	 * decomposed into.
	 * @param i the number of the token bucket
	 * @return the token bucket
	 */
	public ArrivalCurve getTBComponent(int i) {
		if ( token_buckets == null ) {
			decomposeIntoTokenBuckets();
		}
		return token_buckets.get(i);
	}
	
	/**
	 * Returns a string representation of this curve.
	 * @return the curve represented as a string.
	 */
	@Override
	public String toString() {
		String result = "AC{";
		for (int i = 0; i < segments.length; i++) {
			if (i > 0) {
				result += ";";
			}
			if (segments[i].leftopen) {
				result += "!";
			}
			result += "(" + segments[i].x0 + "," + segments[i].y0 + "),"
						+ segments[i].grad;
		}
		result += "}";
		return result;
	}
}
