/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.curves;

import unikl.disco.misc.Num;

/**
 * Class representing linear segments of a curve. A linear segments starts at
 * point (<code>x0</code>, <code>y0</code>) and continues to the right
 * with slope <code>grad</code>. If <code>leftopen</code> is
 * <code>true</code>, the point (<code>x0</code>,<code>y0</code>) is
 * excluded from the segment, otherwise is is included.
 *
 * @author Frank A. Zdarsky
 */
public class LinearSegment{
	public static final LinearSegment X_AXIS = createHorizontalLine( 0.0 );

	/** The x-coordinate of the linear segment's starting point. */
	public Num x0;

	/** The y-coordinate of the linear segment's starting point. */
	public Num y0;

	/** The gradient of the linear segment */
	public Num grad;

	/** Whether the point (<code>x0</code>, <code>y0</code>) is part of the segment or not. */
	public boolean leftopen;

	/**
	 * The default constructor.
	 */
	private LinearSegment() {
		x0       = Num.ZERO;
		y0       = Num.ZERO;
		grad     = Num.ZERO;
		leftopen = false;
	}

	public static LinearSegment createNullSegment() {
		return LinearSegment.X_AXIS.copy();
	}
	
	public static LinearSegment createHorizontalLine( double y ) {
		LinearSegment segment = new LinearSegment();
		segment.y0       = new Num( y );		
		return segment;
	}
	
	/**
	 * A convenience constructor.
	 * @param x0
	 * @param y0
	 * @param grad
	 * @param leftopen
	 */
	public LinearSegment( Num x0, Num y0, Num grad, boolean leftopen ) {
		this.x0       = x0;
		this.y0       = y0;
		this.grad     = grad;
		this.leftopen = leftopen;
	}

	/**
	 * Returns a copy of this instance.
	 *
	 * @return a copy of this instance.
	 */
	public LinearSegment copy() {
		LinearSegment copy = new LinearSegment();
		if( this.x0 != null ) {
			copy.x0 = this.x0.copy();			
		} else {
			try {
				throw new Exception();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		
		if( this.y0 != null ) {
			copy.y0 = this.y0.copy();			
		} else {
			try {
				throw new Exception();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		
		if( this.grad != null ) {
			copy.grad = this.grad.copy();		
		} else {
			try {
				throw new Exception();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		
		copy.leftopen = this.leftopen;
		return copy;
	}

	/**
	 * Returns the function value of this linear segment at the given
	 * x-coordinate. Note that there is no test whether the function is defined
	 * at this location, but simply returns the the value of the co-linear line.
	 *
	 * @param x the coordinate whose function value shall be returned
	 * @return the function value
	 */
	public Num f( Num x ) {
		return Num.add( Num.mult( Num.sub( x, x0 ), grad ), y0 );
	}

	/**
	 * Returns the x-coordinate at which a co-linear line through this segment
	 * intersects a co-linear line through the segment <code>other</code>.
	 *
	 * @param other the other segment
	 * @return the x-coordinate at which the segments cross or NaN of they are parallel
	 */
	public Num getXIntersectionWith( LinearSegment other ) {
		Num y1 = Num.sub( this.y0, Num.mult( x0, this.grad ) );
		Num y2 = Num.sub( other.y0, Num.mult( other.x0, other.grad ) );
		return Num.div( Num.sub( y2, y1 ), Num.sub( this.grad, other.grad ) ); // returns NaN if lines are parallel
	}

	/** Maximum value inside [a,b).
	 *
	 * @param a
	 * @param b
	 * @return the max value inside [a,b)
	 */
	public Num maxAfter( Num a, Num b ) {
		return Num.max( f( Num.max( a, x0 ) ), f( b ) );
	}

	Num getX() { return x0; }
	
	public Num getXValue() { return x0; }
	
	public Num getY() { return y0; }
	
	public Num getYValue() { return y0; }
	
	public void setY( Num y ) { this.y0 = y; }
	
	public boolean equals( LinearSegment other ) {
		boolean result;
		result = this.x0.equals( other.x0 );
		result = result && this.y0.equals( other.y0 );
		result = result && this.grad.equals( other.grad );
		result = result && ( this.leftopen == other.leftopen );
		return result;
	}
}
