/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.curves;

import java.util.ArrayList;
import java.util.List;

import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;

public class ServiceCurve extends Curve {
	public static final ServiceCurve NULL_SERVICE_CURVE = new ServiceCurve( Curve.NULL_CURVE );
	
	protected List<ServiceCurve> rate_latencies;
	
	public ServiceCurve() {
		super();
		rate_latencies = null;
	}

	public ServiceCurve( int segment_count ) {
		super( segment_count );
		rate_latencies = null;
	}
	
	public ServiceCurve( Curve curve ) {
		super( curve.getSegmentCount() );	

		if( Configuration.performChecks() && !this.isWideSenseIncreasing() ) {
			throw new RuntimeException("Service curves can only be created from wide sense increasing curves.");
		}		

		rate_latencies = null;
		alias = curve.getAlias();
		label = curve.getLabel();
		
		for ( int i = 0; i < curve.getSegmentCount(); i++ ) {
			segments[i] = curve.getSegment( i ).copy();
		}
	}
	
	/**
	 * Creates a null arrival curve.
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createNullService() {
		return NULL_SERVICE_CURVE.copy();
	}
	
	/**
	 * Creates a burst curve with zero delay.
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createZeroDelayBurst() {
		return new ServiceCurve( ZERO_DELAY_BURST_CURVE );
	}

	/**
	 * Creates a burst delay curve.
	 * @param t the delay, which must be >= 0.0
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createBurstDelay( Num t ) {
		return new ServiceCurve( Curve.createBurstDelay( t ) );
	}
	
	/**
	 * Creates a new rate latency curve.
	 * @param r the rate
	 * @param t the latency
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createRateLatency( Num r, Num t ) {
		return new ServiceCurve( Curve.createRateLatency( r, t ) );
	}

	/**
	 * Creates a new rate latency curve.
	 * @param r the rate
	 * @param t the latency
	 * @return a <code>Curve</code> instance
	 */
	public static ServiceCurve createRateLatency( double r, double t ) {
		return new ServiceCurve( Curve.createRateLatency( new Num( r ), new Num( t ) ) );
	}

	public ServiceCurve copy() {
		return new ServiceCurve( this );
	}
	
	public static ServiceCurve add( ServiceCurve c1, ServiceCurve c2 ) {
		return new ServiceCurve( Curve.add(c1, c2) );
	}
	
	public static ServiceCurve sub( ServiceCurve c1, ArrivalCurve c2 ) {
		return new ServiceCurve( Curve.sub(c1, c2) );
	}
	
	public static ServiceCurve min( ServiceCurve c1, ServiceCurve c2 ) {
		return new ServiceCurve( Curve.min(c1, c2) );
	}
	
	public static ServiceCurve max( ServiceCurve c1, ServiceCurve c2 ) {
		return new ServiceCurve( Curve.max(c1, c2) );
	}
	
	/**
	 * Returns a copy of curve bounded at the x-axis.
	 * @return the bounded curve.
	 */
	public static ServiceCurve boundAtXAxis( ServiceCurve service_curve ) {
		return new ServiceCurve( Curve.boundAtXAxis( service_curve ) ); 
	}
	
	/**
	 * Decomposes this curve into a list of rate latency curves and
	 * stores this list in the curve's <code>rate_latencies</code> field.<br/>
	 * Note: Curve must be convex.
	 */
	private void decomposeIntoRateLatencies() {
		if( Configuration.performChecks() && !this.isConvex() ) {
			if ( this.equals( ServiceCurve.createZeroDelayBurst() ) ) {
				rate_latencies = new ArrayList<ServiceCurve>();
				rate_latencies.add( createRateLatency( Num.POSITIVE_INFINITY, Num.ZERO ) );
			} else {
				throw new RuntimeException("Can only decompose convex service curves into rate latency curves.");
			}
		} else {
			rate_latencies = new ArrayList<ServiceCurve>();
			for (int i = 0; i < segments.length; i++) {
				if (segments[i].y0.equals( 0.0 ) && segments[i].grad.equals( 0.0 ) ) {
					continue;
				}
				Num r = segments[i].grad;
				Num l = Num.sub( segments[i].x0, Num.div( segments[i].y0, segments[i].grad ) );
				rate_latencies.add( createRateLatency( r, l ) );
			}
		}
	}

	/**
	 * Returns the number of rate latency curves the curve can be decomposed into.
	 * @return the number of rate latency curves
	 */
	public int getRLComponentCount() {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies.size();
	}

	/**
	 * Returns a list of rate latency curves that this curve can be
	 * decomposed into.
	 * @return the list of rate latency curves
	 */
	public final List<ServiceCurve> getRLComponents() {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies;
	}

	/**
	 * Returns the <code>i</code>th rate latency curve that this curve can be
	 * decomposed into.
	 * @param i the number of the rate latency curve
	 * @return the rate latency curve
	 */
	public ServiceCurve getRLComponent(int i) {
		if (rate_latencies == null) {
			decomposeIntoRateLatencies();
		}
		return rate_latencies.get( i );
	}

	/**
	 * Returns a copy of this curve with latency removed, i.e. shifted left by
	 * the latency.
	 * @return a copy of this curve without latency
	 */
	public ServiceCurve removeLatency() {
		return new ServiceCurve( Curve.removeLatency( this ) );
	}
	
	public boolean equals( ServiceCurve other ) {
		return ( (Curve)this ).equals( other );
	}
	
	/**
	 * Returns a string representation of this curve.
	 * @return the curve represented as a string.
	 */
	@Override
	public String toString() {
		String result = "SC{";
		for (int i = 0; i < segments.length; i++) {
			if (i > 0) {
				result += ";";
			}
			if (segments[i].leftopen) {
				result += "!";
			}
			result += "(" + segments[i].x0 + "," + segments[i].y0 + "),"
						+ segments[i].grad;
		}
		result += "}";
		return result;
	}
}
