/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *                              disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.minplus;

import unikl.disco.curves.ArrivalCurve;
import unikl.disco.curves.Curve;
import unikl.disco.curves.LinearSegment;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;

public class Convolution {
	public static ArrivalCurve convolve( ArrivalCurve arrival_curve, ServiceCurve service_curve ) {
		if ( service_curve.getLatency().equals( Num.POSITIVE_INFINITY ) ) {
			return ArrivalCurve.createNullArrival();
		}

		ArrivalCurve result = new ArrivalCurve( Curve.min( service_curve.removeLatency(), arrival_curve ) );
		result.shiftRight( service_curve.getLatency() );
		
		result.beautify();
		
		if ( Configuration.createLabels()
				&& ( arrival_curve.getLabel() != null || arrival_curve.getAlias() != null
						|| service_curve.getLabel() != null || service_curve.getAlias() != null) ) {
			String label1 = ( arrival_curve.getAlias().length() > 0 ) ? arrival_curve.getAlias() : "?";
			String label2 = ( service_curve.getAlias().length() > 0 ) ? service_curve.getAlias() : "?";
			result.setLabel( "(" + label1 + "\u2297" + label2 + ")" );
		}
		return result;
	}
	
	/**
	 * Returns the convolution of two curve, which must be convex
	 * 
	 * @param service_curve_1 first the curve to convolve with.
	 * @param service_curve_2 second the curve to convolve with.
	 * @return the convolved curve.
	 */
	public static ServiceCurve convolve( ServiceCurve service_curve_1, ServiceCurve service_curve_2 ) {
		if ( service_curve_1.equals( ServiceCurve.ZERO_DELAY_BURST_CURVE ) ) {
			return service_curve_2.copy();
		} else {
			if ( service_curve_2.equals( ServiceCurve.ZERO_DELAY_BURST_CURVE ) ) {
				return service_curve_1.copy();
			}
		}
		
		if ( service_curve_1.isBurstDelay() && service_curve_1.getSegment(1).x0.equals( Num.ZERO ) ) {
			return service_curve_2.copy();
		}
		if ( service_curve_2.isBurstDelay() && service_curve_2.getSegment(1).x0.equals( Num.ZERO ) ) {
			return service_curve_1.copy();
		}

		ServiceCurve result = new ServiceCurve();

		LinearSegment s;
		s = LinearSegment.createHorizontalLine( 0.0 );
		s.x0 = Num.ZERO;
		s.y0 = Num.add( service_curve_1.f( Num.ZERO ), service_curve_2.f( Num.ZERO ) );
		result.addSegment(s);

		int i1 = (service_curve_1.isRealDiscontinuity(0)) ? 1 : 0;
		int i2 = (service_curve_2.isRealDiscontinuity(0)) ? 1 : 0;
		if (i1 > 0 || i2 > 0) {
			s = LinearSegment.createHorizontalLine( 0.0 );
			s.x0       = Num.ZERO;
			s.y0       = Num.add( service_curve_1.fLimitRight( Num.ZERO ), service_curve_2.fLimitRight( Num.ZERO ) );
			s.leftopen = true;
			result.addSegment(s);
		}

		while(i1 < service_curve_1.getSegmentCount() || i2 < service_curve_2.getSegmentCount()) {
			if ( service_curve_1.getSegment(i1).grad.less( service_curve_2.getSegment(i2).grad ) ) {
				if (i1+1 >= service_curve_1.getSegmentCount()) {
					result.getSegment(result.getSegmentCount()-1).grad = service_curve_1.getSegment(i1).grad;
					break;
				}

				s = LinearSegment.createHorizontalLine( 0.0 );
				s.x0 = Num.add( result.getSegment( result.getSegmentCount()-1 ).x0,
						( Num.sub( service_curve_1.getSegment( i1+1 ).x0, service_curve_1.getSegment( i1 ).x0 ) ) );
				s.y0 = Num.add( result.getSegment( result.getSegmentCount()-1 ).y0,
						( Num.sub( service_curve_1.getSegment( i1+1 ).y0, service_curve_1.getSegment( i1 ).y0 ) ) );
				result.getSegment(result.getSegmentCount()-1).grad = service_curve_1.getSegment(i1).grad;
				result.addSegment(s);
				i1++;
			} else {
				if (i2+1 >= service_curve_2.getSegmentCount()) {
					result.getSegment(result.getSegmentCount()-1).grad = service_curve_2.getSegment(i2).grad;
					break;
				}

				s = LinearSegment.createHorizontalLine( 0.0 );
				s.x0 = Num.add( result.getSegment( result.getSegmentCount()-1 ).x0,
						( Num.sub( service_curve_2.getSegment( i2+1 ).x0, service_curve_2.getSegment( i2 ).x0 ) ) );
				s.y0 = Num.add( result.getSegment( result.getSegmentCount()-1 ).y0,
						( Num.sub( service_curve_2.getSegment( i2+1 ).y0, service_curve_2.getSegment( i2 ).y0 ) ) );
				result.getSegment(result.getSegmentCount()-1).grad = service_curve_2.getSegment(i2).grad;
				result.addSegment(s);
				i2++;
			}
		}
		
		result.beautify();
		
		if ( Configuration.createLabels()
				&& (service_curve_1.getLabel() != null || service_curve_1.getAlias() != null
						|| service_curve_2.getLabel() != null || service_curve_2.getAlias() != null)) {
			String label1 = (service_curve_1.getAlias().length() > 0) ? service_curve_1.getAlias() : "?";
			String label2 = (service_curve_2.getAlias().length() > 0) ? service_curve_2.getAlias() : "?";
			result.setLabel ( "(" + label1 + "\u2297" + label2 + ")" );
		}
		return result;
	}
}
