/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *                              disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.minplus;

import unikl.disco.curves.ArrivalCurve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.misc.Num;
import unikl.disco.nc.Configuration;

public class Deconvolution {
	/**
	 * Returns the deconvolution of an (almost) concave arrival curve and
	 * a convex service curve.
	 * 
	 * @param arrival_curve the (almost) concave arrival curve.
	 * @param service_curve the convex service curve.
	 * @return the deconvolved curve, an arrival curve.
	 */
	public static ArrivalCurve deconvolve( ArrivalCurve arrival_curve, ServiceCurve service_curve ) {
		if( service_curve.equals( ServiceCurve.ZERO_DELAY_BURST_CURVE ) ) {
			return arrival_curve;
		}
		if( arrival_curve.equals( ArrivalCurve.NULL_CURVE ) ) {
			return arrival_curve;
		}
		if( service_curve.equals( ServiceCurve.NULL_CURVE )
				|| service_curve.getLatency().equals( Num.POSITIVE_INFINITY )
				|| ( service_curve.getSustainedRate().equals( Num.ZERO ) && service_curve.getSegment( service_curve.getSegmentCount() - 1 ).y0.equals( Num.ZERO ) ) ) {
			return ArrivalCurve.createNullArrival();
		}
		if ( Configuration.performChecks() && !arrival_curve.isAlmostConcave() ) {
			throw new IllegalArgumentException("Arrival curve of deconvolution must be almost concave.");
		}
		if ( Configuration.performChecks() && !service_curve.isConvex() ) {
			throw new IllegalArgumentException("Service curve of deconvolution must be convex.");
		}
		
		int i_this  = 0;
		int i_other = 0;

		// advance this_iterator to concave part
		while(i_this < arrival_curve.getSegmentCount()) {
			if ( arrival_curve.getSegment( i_this ).grad.greater( Num.ZERO ) ) {
				break;
			}
			i_this++;
		}
		// advance other_iterator to first IP after the start of this's concave part
		while(i_other < service_curve.getSegmentCount()) {
			if ( service_curve.getSegment( i_other ).x0.ge( arrival_curve.getSegment( i_this ).x0 ) ) {
				break;
			}
			i_other++;
		}

		// from here on, search both curve's IPs until grad(other)>=grad(this)
		Num kj = Num.POSITIVE_INFINITY;
		while(i_this < arrival_curve.getSegmentCount() || i_other < service_curve.getSegmentCount()) {
			Num x_this = ( i_this < arrival_curve.getSegmentCount() ) ?
					arrival_curve.getSegment( i_this ).x0 : Num.POSITIVE_INFINITY;
			Num x_other = (i_other < service_curve.getSegmentCount()) ?
					service_curve.getSegment(i_other).x0 : Num.POSITIVE_INFINITY;

			if ( x_other.le( x_this ) ) {
				if ( service_curve.getSegment(i_other).grad.ge
						( arrival_curve.getGradientLimitRight( service_curve.getSegment( i_other ).x0) ) ) {
					kj = service_curve.getSegment(i_other).x0;
					break;
				}
				i_other++;
			} else {
				if ( service_curve.getGradientLimitRight( arrival_curve.getSegment( i_this ).x0 ).ge
						( arrival_curve.getSegment( i_this ).grad ) ) {
					kj = arrival_curve.getSegment(i_this).x0;
					break;
				}
				i_this++;
			}
		}

		if ( kj.less( Num.POSITIVE_INFINITY ) ) {
			ArrivalCurve result = ArrivalCurve.add( arrival_curve.shiftLeftClipping( kj ), Num.negate( service_curve.f( kj ) ) );
			
			result.beautify();
			
			if ( Configuration.createLabels()
					&& (arrival_curve.getLabel() != null || arrival_curve.getAlias() != null
							|| service_curve.getLabel() != null || service_curve.getAlias() != null)) {
				String label1 = (arrival_curve.getAlias().length() > 0) ? arrival_curve.getAlias() : "?";
				String label2 = (service_curve.getAlias().length() > 0) ? service_curve.getAlias() : "?";
				result.setLabel ( "(" + label1 + "\u2298" + label2 + ")" );
			}
			return result;
		} else {
			ArrivalCurve c = new ArrivalCurve( 1 );
			c.getSegment(0).y0 = Num.POSITIVE_INFINITY;
			return c;
		}
	}
}
