/*
 *  (c) 2013 - 2014 Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.ArrivalBoundFlags;
import unikl.disco.nc.Configuration.GammaFlag;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Server;
import unikl.disco.network.Link;

public abstract class ArrivalBound {
	protected Network network;
	
	static Cache ab_cache = new Cache();
	
	protected abstract ArrivalCurve computeArrivalBound( Server server ) throws Exception;
	
	protected abstract ArrivalCurve computeArrivalBound( Server server, Set<Flow> flows_to_bound ) throws Exception;
	
	protected abstract ArrivalCurve computeArrivalBound( Server server, Set<Flow> flows_to_bound, Flow flow_of_interest ) throws Exception;
	
	protected abstract ArrivalCurve computeArrivalBound( Link link ) throws Exception;
	
	protected abstract ArrivalCurve computeArrivalBound( Link link, Set<Flow> flows_to_bound ) throws Exception;
	
	protected abstract ArrivalCurve computeArrivalBound( Link link, Set<Flow> flows_to_bound, Flow flow_of_interest ) throws Exception;
	
	public static ArrivalCurve computeArrivalBound( Network network, Server server ) throws Exception {
		return computeArrivalBound( network, server, network.getFlows( server ), Flow.NULL_FLOW );
	}
	
	public static ArrivalCurve computeArrivalBound( Network network, Server server, Set<Flow> flows_to_bound ) throws Exception {
		return computeArrivalBound( network, server, flows_to_bound, Flow.NULL_FLOW );
	}
	
	public static ArrivalCurve computeArrivalBound( Network network, Link link ) throws Exception {
		if( !link.getSource().equals( link.getDest() ) ) {
			throw new Exception( "link does not end in the given server" );
		}
		return computeArrivalBound( network, link.getDest(), network.getFlows( link ), Flow.NULL_FLOW );
	}
	
	public static ArrivalCurve computeArrivalBound( Network network, Link link, Set<Flow> flows_to_bound ) throws Exception {
		if( !link.getSource().equals( link.getDest() ) ) {
			throw new Exception( "link does not end in the given server" );
		}
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( flows_to_bound, network.getFlows( link ) );
		return computeArrivalBound( network, link.getDest(), flows_on_l_to_bound, Flow.NULL_FLOW );
	}
	
	public static ArrivalCurve computeArrivalBound( Network network, Link link, Set<Flow> flows_to_bound, Flow flow_of_interest ) throws Exception {
		if( !link.getSource().equals( link.getDest() ) ) {
			throw new Exception( "link does not end in the given server" );
		}
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( flows_to_bound, network.getFlows( link ) );
		return computeArrivalBound( network, link.getDest(), flows_on_l_to_bound, Flow.NULL_FLOW );
	}
	
	/**
	 * Calculates the arrival curve for a the flows in flows_to_bound that arrive at the given server
	 * while considering the lower priority of the flow_of_interest
	 * 
	 * @param network
	 * @param server
	 * @param flows_to_bound
	 * @return the arrival bound at the specified server
	 * @throws Exception
	 */
	public static ArrivalCurve computeArrivalBound( Network network, Server server, Set<Flow> flows_to_bound, Flow flow_of_interest ) throws Exception {
		if( Configuration.useCache() && Configuration.multiplexingDiscipline() != MuxDiscipline.SERVERLOCAL_FLAG ) { // Do not cache in that case. Too many variables
			CacheEntry entry = ab_cache.getEntry( server, flows_to_bound, flow_of_interest );
			if ( entry != null ) {
				return entry.arrival_curve;
			}			
		}
		
		ArrivalCurve arrival_bound = ArrivalCurve.NULL_ARRIVAL_CURVE;

		switch( Configuration.arrivalBoundCalculation() ) {
			case PMOO_BOUNDING:
				PmooArrivalBound pmoo_arrival_bound = new PmooArrivalBound( network );
				arrival_bound = pmoo_arrival_bound.computeArrivalBound( server, flows_to_bound, flow_of_interest );
			break;
			
			case PER_HOP_BOUNDING:
			default:
				PerHopArrivalBound per_hop_bound = new PerHopArrivalBound( network );
				arrival_bound = per_hop_bound.computeArrivalBound( server, flows_to_bound, flow_of_interest );
			break;
		}
	
	arrival_bound.beautify();

	if( Configuration.useCache() ) {
		ab_cache.addEntry( server, flows_to_bound, flow_of_interest, arrival_bound );		
	}
	return arrival_bound;
	}
}

//There's also the additional possibility to create classes of flows_to_bound that share equivalent cross-traffic
//and thus can reuse each other's results (see test documentation, e.g., FeedForward_1SC_4Flows_1AC_4Paths).
//However, it is necessary to analyze the topology to find the possible equivalence classes.
//This is considered too expensive.
class Cache {	
	Map<Server,Set<CacheEntry>> map__server__entries = new HashMap<Server,Set<CacheEntry>>();
	Map<Flow,Set<CacheEntry>> map__foi__entries = new HashMap<Flow,Set<CacheEntry>>();
	
	// No checks for existing entry with the same variables!
	protected void addEntry( Server server, Set<Flow> flows_to_bound, Flow flow_of_interest, ArrivalCurve arrival_curve ) {
		CacheEntry entry = new CacheEntry( server, flows_to_bound, flow_of_interest, arrival_curve );

		Set<CacheEntry> entries_s = map__server__entries.get( server );
		Set<CacheEntry> entries_f = map__foi__entries.get( flow_of_interest );
				
		if( entries_s == null ) {
			map__server__entries.put( server, new HashSet<CacheEntry>( Collections.singleton( entry ) ) );
		} else {
			entries_s.add( entry );
		}
		
		if( entries_f == null ) {
			map__foi__entries.put( flow_of_interest, new HashSet<CacheEntry>( Collections.singleton( entry ) ) );
		} else {
			entries_f.add( entry );
		}
	}
	
	protected CacheEntry getEntry( Server server, Set<Flow> flows_to_bound, Flow flow_of_interest ) {
		// Most important feature is an efficient search
		Set<CacheEntry> entries_s = map__server__entries.get( server );
		Set<CacheEntry> entries_f = map__foi__entries.get( flow_of_interest );

		Set<CacheEntry> candidates = SetUtils.getIntersection( entries_s, entries_f );
		
		for ( CacheEntry entry : candidates ) {
			if( entry.flows_to_bound.size() == flows_to_bound.size()
					&& entry.ab_method == Configuration.arrivalBoundCalculation()
					&& entry.mux == Configuration.multiplexingDiscipline()
					&& entry.gamma == Configuration.useGamma()
					&& entry.extra_gamma == Configuration.useExtraGamma() ) {
				
				// should be the most expensive operation so do it last
				if( entry.flows_to_bound.equals( flows_to_bound ) ) {
					return entry;					
				}
			}
		}
		
		return null;
	}

	protected void invalidateCache() {
		map__server__entries = new HashMap<Server,Set<CacheEntry>>();
		map__foi__entries = new HashMap<Flow,Set<CacheEntry>>();
	}
}

class CacheEntry {
	Server server;
	Set<Flow> flows_to_bound;
	Flow flow_of_interest;
	
	ArrivalBoundFlags ab_method;
	MuxDiscipline mux;
	GammaFlag gamma;
	GammaFlag extra_gamma;
	
	ArrivalCurve arrival_curve;
	
	CacheEntry( Server server, Set<Flow> flows_to_bound, Flow flow_of_interest, ArrivalCurve arrival_curve ) {
		this.server = server; 
		this.flows_to_bound = flows_to_bound;
		this.flow_of_interest = flow_of_interest;
		
		this.ab_method = Configuration.arrivalBoundCalculation();
		this.mux = Configuration.multiplexingDiscipline();
		this.gamma = Configuration.useGamma();
		this.extra_gamma = Configuration.useExtraGamma();
		
		this.arrival_curve = arrival_curve;
	}
}