/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.List;

import unikl.disco.curves.Curve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.Num;

public final class LeftOverService {
	private LeftOverService() {}
	
	/**
	 * Computes the left-over FIFO service curve for a server with the service curve
	 * <code>beta</code> experiencing cross-traffic with arrival curve <code>alpha</code>.
	 * 
	 * @param arrival_curve arrival curve of cross-traffic
	 * @param service_curve server's service curve
	 * @return the FIFO service curve
	 */
	public static ServiceCurve fifoMux( ServiceCurve service_curve, ArrivalCurve arrival_curve ) {
		if ( Configuration.performChecks() && !arrival_curve.isConcave() ) {
			throw new IllegalArgumentException( "Arrival curve must be concave." );
		}
		if ( Configuration.performChecks() && !service_curve.isConvex() ) {
			throw new IllegalArgumentException( "Service curve must be convex." );
		}

		Num last_y = arrival_curve.fLimitRight( Num.ZERO );

		List<Num> ycoords = Curve.computeInflectionPointsY( arrival_curve, service_curve );
		for( int i = 0; i < ycoords.size(); i++ ) {
			Num ip_y = ( ycoords.get( i ) );
			if ( ip_y.less( last_y ) ) {
				continue;
			}

			Num x_alpha = arrival_curve.f_inv( ip_y, false );
			Num x_beta  = service_curve.f_inv( ip_y, true );
			if ( arrival_curve.getGradientLimitRight( x_alpha ).le( service_curve.getGradientLimitRight( x_beta ) ) ) {
				
				Num theta = Num.sub( x_beta, x_alpha );
				ServiceCurve beta_fifo = ServiceCurve.boundAtXAxis(
											ServiceCurve.min( 
												ServiceCurve.sub( service_curve, arrival_curve.shiftRight( theta ) ), 
												ServiceCurve.createBurstDelay( x_beta )
											) 
										);
	    		return beta_fifo;
			}
		}
		return service_curve.copy();
	}
	
	/**
	 * Computes the left-over service curve for a server under arbitrary multiplexing
	 * with the service curve <code>beta</code> experiencing cross-traffic 
	 * with arrival curve <code>alpha</code>.
	 * 
	 * @param alpha arrival curve of cross-traffic
	 * @param beta server's service curve
	 * @return the FIFO service curve
	 */
	public static ServiceCurve arbMux( ServiceCurve beta, ArrivalCurve alpha ) {
		return new ServiceCurve( ServiceCurve.boundAtXAxis( ServiceCurve.sub( beta, alpha ) ) );
	}
}
