/*
 *  (c) 2013 - 2014 Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.Stack;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.nc.Configuration.GammaFlag;
import unikl.disco.network.Flow;
import unikl.disco.network.Link;
import unikl.disco.network.Network;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;
import unikl.disco.minplus.Deconvolution;

public class PerHopArrivalBound extends ArrivalBound {
	@SuppressWarnings("unused")
	private PerHopArrivalBound() {}

	public PerHopArrivalBound( Network network ) {
		super.network = network;
	}

	protected ArrivalCurve computeArrivalBound( Server server ) throws Exception {
		return computeArrivalBound( server, network.getFlows( server ), Flow.NULL_FLOW );
	}
	
	protected ArrivalCurve computeArrivalBound( Server server, Set<Flow> f_xfcaller ) throws Exception {
		return computeArrivalBound( server, f_xfcaller, Flow.NULL_FLOW );
	}

	protected ArrivalCurve computeArrivalBound( Link link ) throws Exception {
		return computeArrivalBound( link.getSource(), network.getFlows( link ), Flow.NULL_FLOW );
	}
	
	protected ArrivalCurve computeArrivalBound( Link link, Set<Flow> f_xfcaller ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( f_xfcaller, network.getFlows( link ) );
		return computeArrivalBound( link.getSource(), flows_on_l_to_bound, Flow.NULL_FLOW );
	}

	protected ArrivalCurve computeArrivalBound( Link link, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( f_xfcaller, network.getFlows( link ) );
		return computeArrivalBound( link.getSource(), flows_on_l_to_bound, flow_of_interest );
	}
	
	/**
	 * Returns the output bound for a set of <code>f_xfcaller</code> from
	 * the given <code>server</code>.<br/>
	 * It is possible to change between iterative and recursive computation by
	 * calling <code>setUseRecursiveOutputBoundComputation()</code>.
	 * 
	 * @param server
	 *            the server for which the arrival bound is computed
	 * @param f_xfcaller
	 *            set of flows for which the arrival bound is computed
	 * @param flow_of_interest
	 *            the flow with the lowest priority
	 * @return a <code>Curve</code> instance containing the output bound
	 * @throws OutOfMemoryError
	 * @author Andreas Kiefer
	 */
	protected ArrivalCurve computeArrivalBound( Server server, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws Exception
	{
		ArrivalCurve arrival_bound = ArrivalCurve.createNullArrival();
		
		// Get flows to bound that are actually present at server
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_xfcaller, network.getFlows( server ) );
		f_xfcaller_server.remove( flow_of_interest );
		if ( f_xfcaller_server.size() == 0 )
		{
			return arrival_bound;
		}
		
		// Get cross-traffic originating in server
		Set<Flow> f_xfcaller_sourceflows_server = SetUtils.getIntersection( f_xfcaller_server, network.getSourceFlows( server ) );
		arrival_bound = network.getSourceFlowArrivalCurve( server, f_xfcaller_sourceflows_server );
		
		// Get cross-traffic from each predecessor
		ArrivalCurve arrival_bound_predec = ArrivalCurve.NULL_ARRIVAL_CURVE;
		Set<Flow> f_xfcaller_server_bounded = f_xfcaller_sourceflows_server;
		for ( Link in_l : network.getInLinks( server ) )
		{
			Set<Flow> f_xfcaller_predec = SetUtils.getIntersection( network.getFlows( in_l ), f_xfcaller_server );
			
			if( f_xfcaller_predec.isEmpty() ) { // Do not check links without flows of interest
				continue;
			}
			
			if ( Configuration.useRecursiveArrivalBoundComputation() )
			{
				arrival_bound_predec = computeRecursiveOutputBound( in_l.getSource(), f_xfcaller_predec, flow_of_interest );
			}
			else
			{
				arrival_bound_predec = computeIterativeOutputBound( in_l.getSource(), f_xfcaller_predec, flow_of_interest );
			}
			arrival_bound = ArrivalCurve.add( arrival_bound, arrival_bound_predec );
			
			f_xfcaller_server_bounded.addAll( f_xfcaller_predec ); // Stop as soon as all flows are bounded
			if ( f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
				break;
			}
		}
		
		return arrival_bound;
	}

	/**
	 * Returns the arrival bound for a set of <code>f_xfcaller</code> from
	 * the given <code>server</code> by recursion.<br/>
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param f_xfcaller
	 *            set of flows for which the output bound is computed
	 * @param flow_of_interest
	 *            the flow with the lowest priority
	 * @return a <code>Curve</code> instance containing the output bound
	 * @throws OutOfMemoryError
	 */
	protected ArrivalCurve computeRecursiveOutputBound( Server server, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws OutOfMemoryError
	{
		ArrivalCurve alpha_xfcaller = ArrivalCurve.createNullArrival();
		f_xfcaller.remove( flow_of_interest );

		if ( f_xfcaller.size() == 0 )
		{
			return alpha_xfcaller;
		}
		
		// Get flows originating in this server and sort by 
		// flows that are cross traffic at the next server (f_xfcaller_server, i.e, cross traffic of the caller)
		// flow that are cross traffic to them (f_xxfcaller_server, i.e., cross traffic to the cross traffic of the caller)
		Set<Flow> f_server = network.getFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		
		if( Configuration.useCache() && Configuration.multiplexingDiscipline() != MuxDiscipline.SERVERLOCAL_FLAG ) { // Do not cache in that case. Too many variables
			try {
				// need succeeding server due to this method's output bound point of view
				CacheEntry entry = ab_cache.getEntry( flow_of_interest.getSucceedingServer( server ), f_xfcaller_server, flow_of_interest );
				if ( entry != null ) {
					return entry.arrival_curve;
				}
			} catch (Exception e) {}
		}
		
		Set<Flow> f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
		f_xxfcaller_server.remove( flow_of_interest );

		ArrivalCurve alpha_xfcaller_link_sum = ArrivalCurve.NULL_ARRIVAL_CURVE;
		ArrivalCurve alpha_xxfcaller_link_sum = ArrivalCurve.NULL_ARRIVAL_CURVE;
		
		Set<Flow> f_link;
		Set<Flow> f_xfcaller_link;
		Set<Flow> f_xxfcaller_link;
		
		if( Configuration.useCache() && Configuration.multiplexingDiscipline() != MuxDiscipline.SERVERLOCAL_FLAG ) { // Do not cache in that case. Too many variables			
			for ( Link in_l : network.getInLinks( server ) )
			{
				ArrivalCurve alpha_xfcaller_link = ArrivalCurve.NULL_ARRIVAL_CURVE;
				ArrivalCurve alpha_xxfcaller_link = ArrivalCurve.NULL_ARRIVAL_CURVE;
								
				f_link = network.getFlows( in_l );
				f_link.remove( flow_of_interest ); // flow of interest has the lowest priority
				
				f_xfcaller_link = SetUtils.getIntersection( f_link, f_xfcaller_server );
				f_xxfcaller_link = SetUtils.getDifference( f_link, f_xfcaller_link );
				
				CacheEntry xf_entry = ab_cache.getEntry( server, f_xfcaller_link, flow_of_interest );
				CacheEntry xxf_entry = ab_cache.getEntry( server, f_xxfcaller_link, flow_of_interest );
				
				if ( xf_entry != null ) {
					alpha_xfcaller_link_sum = ArrivalCurve.add( alpha_xfcaller_link_sum, xf_entry.arrival_curve );
				} else {
					alpha_xfcaller_link = computeRecursiveOutputBound( in_l.getSource(), f_xfcaller_link, flow_of_interest );
					alpha_xfcaller_link_sum = ArrivalCurve.add( alpha_xfcaller_link_sum, alpha_xfcaller_link );
					ArrivalBound.ab_cache.addEntry( server, f_xfcaller_link, flow_of_interest, alpha_xfcaller_link );
				}
				
				if ( xxf_entry != null ) {
					alpha_xxfcaller_link_sum = ArrivalCurve.add( alpha_xxfcaller_link_sum, xxf_entry.arrival_curve );
				} else {
					alpha_xxfcaller_link = computeRecursiveOutputBound( in_l.getSource(), f_xxfcaller_link, flow_of_interest );
					alpha_xxfcaller_link_sum = ArrivalCurve.add( alpha_xxfcaller_link_sum, alpha_xxfcaller_link );
					ArrivalBound.ab_cache.addEntry( server, f_xxfcaller_link, flow_of_interest, alpha_xxfcaller_link );
				}
			}
		} else {
			for ( Link in_l : network.getInLinks( server ) )
			{
				f_link = network.getFlows( in_l );
				f_link.remove( flow_of_interest ); // flow of interest has the lowest priority
				
				f_xfcaller_link = SetUtils.getIntersection( f_link, f_xfcaller_server );
				f_xxfcaller_link = SetUtils.getDifference( f_link, f_xfcaller_link );
				
				alpha_xfcaller_link_sum = ArrivalCurve.add( alpha_xfcaller_link_sum, computeRecursiveOutputBound( in_l.getSource(), f_xfcaller_link, flow_of_interest ) );
				alpha_xxfcaller_link_sum = ArrivalCurve.add( alpha_xxfcaller_link_sum, computeRecursiveOutputBound( in_l.getSource(), f_xxfcaller_link, flow_of_interest ) );
			}
		}

		ArrivalCurve alpha_xfcaller_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_server );
		ArrivalCurve alpha_xxfcaller_server = network.getSourceFlowArrivalCurve( server, f_xxfcaller_server );

		alpha_xfcaller_server = ArrivalCurve.add( alpha_xfcaller_server, alpha_xfcaller_link_sum );
		alpha_xxfcaller_server = ArrivalCurve.add( alpha_xxfcaller_server, alpha_xxfcaller_link_sum );
		
		alpha_xfcaller_server.beautify();
		alpha_xxfcaller_server.beautify();

		ServiceCurve beta_loxfcaller_server = server.getServiceCurve();
		ServiceCurve gamma = server.getMaxServiceCurve();
		if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_ARBITRARY
			|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVERLOCAL_FLAG & server.useArbitraryMultiplexing() ) )
		{
			beta_loxfcaller_server = LeftOverService.arbMux( beta_loxfcaller_server, alpha_xxfcaller_server );
		}
		else
		{
			beta_loxfcaller_server = LeftOverService.fifoMux( beta_loxfcaller_server, alpha_xxfcaller_server );
		}

		if( Configuration.useGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useGamma() == GammaFlag.SERVERLOCAL_FLAG & server.useGamma() ) )
		{
			alpha_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alpha_xfcaller_server, gamma ), beta_loxfcaller_server );
		}
		else
		{
			alpha_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_server, beta_loxfcaller_server );
		}
		
		if( Configuration.useExtraGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useExtraGamma() == GammaFlag.SERVERLOCAL_FLAG & server.useExtraGamma() ) )
		{
			alpha_xfcaller = Convolution.convolve( alpha_xfcaller, gamma.removeLatency() );
		}
		
		alpha_xfcaller.beautify();
		
		if( Configuration.useCache() && Configuration.multiplexingDiscipline() != MuxDiscipline.SERVERLOCAL_FLAG ) {	
			ab_cache.addEntry( server, f_xfcaller_server, flow_of_interest, alpha_xfcaller_server );
			ab_cache.addEntry( server, f_xxfcaller_server, flow_of_interest, alpha_xxfcaller_server );
			try {
				ab_cache.addEntry( flow_of_interest.getSucceedingServer( server ), f_xfcaller_server, flow_of_interest, alpha_xfcaller );
			} catch (Exception e) {}
		}
		
		return alpha_xfcaller;
	}

	/**
	 * Returns the output bound for a set of <code>f_xfcaller</code> from
	 * the given <code>server</code> by iteration.<br/>
	 * For the iteration the help of a stack is required, to track the tuples of
	 * <code>server</code> and <code>f_xfcaller</code>. Also a table is
	 * needed for mapping the tuple to the result.
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param f_xfcaller
	 *            set of flows for which the output bound is computed
	 * @param flow_of_interest
	 *            the flow with the lowest priority
	 * @return a <code>Curve</code> instance containing the output bound
	 * @throws OutOfMemoryError
	 * @author Andreas Kiefer
	 */
	protected ArrivalCurve computeIterativeOutputBound( Server server, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws OutOfMemoryError
	{
		Stack<Pair<Server, Set<Flow>>> stack = new Stack<Pair<Server, Set<Flow>>>();
		Map<Pair<Server, Set<Flow>>, ArrivalCurve> results = new HashMap<Pair<Server, Set<Flow>>, ArrivalCurve>();

		Pair<Server, Set<Flow>> next = new Pair<Server, Set<Flow>>( server, f_xfcaller );
		stack.push( next );
		// until everything is computed
		while ( !stack.isEmpty() )
		{
			if ( !decrementOutputBoundStack( stack, results, flow_of_interest ) )
			{
				incrementOutputBoundStack( stack, results, flow_of_interest );
			}
		}
		ArrivalCurve alpha_xfcaller = results.get( next );
		if ( alpha_xfcaller == null )
		{
			alpha_xfcaller = ArrivalCurve.createNullArrival();
		}

		return alpha_xfcaller;
	}

	/**
	 * Puts all required output bound calculations for a given
	 * <code>server</code> on a <code>stack</code><br/>
	 * When the output bound for a given <code>server</code> and a set of
	 * <code>f_xfcaller</code> cannot be computed, it's because the
	 * calculation depended on results that are not yet computed, so all
	 * dependencies have to be put on the stack. That means, iterating over all
	 * the incoming links of the server.
	 * 
	 * @param stack
	 *            the stack with the <code>server</code>,
	 *            <code>f_xfcaller</code>-tuple.
	 * @param results
	 *            maps the tuple to the result.
	 * @param flow_of_interest
	 *            the flow with the lowest priority
	 * @author Andreas Kiefer
	 */
	protected void incrementOutputBoundStack(	Stack<Pair<Server, Set<Flow>>> stack,
												Map<Pair<Server, Set<Flow>>, ArrivalCurve> results,
												Flow flow_of_interest )
	{
		Pair<Server, Set<Flow>> next = stack.peek();

		Server server = (Server) next.getFirst();
		Set<Flow> f_xfcaller = next.getSecond();
		
		Set<Flow> f_server = network.getSourceFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		f_xfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority
		Set<Flow> f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
		f_xxfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority

		for ( Link in_l : network.getInLinks( server ) )
		{
			f_server = network.getFlows( in_l );
			f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
			f_xfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority
			f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
			f_xxfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority

			// is the required output bound already computed?
			Pair<Server, Set<Flow>> p_incl = new Pair<Server, Set<Flow>>( in_l.getSource(), f_xfcaller_server );
			if ( !results.containsKey( p_incl ) )
			{
				stack.push( p_incl );
				// some output bounds doesn't depend on anything
				if ( calculabe( in_l.getSource(), f_xfcaller_server ) )
				{
					decrementOutputBoundStack( stack, results, flow_of_interest );
				}
			}
			// is the required output bound already computed?
			Pair<Server, Set<Flow>> p_excl = new Pair<Server, Set<Flow>>( in_l.getSource(), f_xxfcaller_server );
			if ( !results.containsKey( p_excl ) )
			{
				stack.push( p_excl );
				// some output bounds doesn't depend on anything
				if ( calculabe( in_l.getSource(), f_xxfcaller_server ) )
				{
					decrementOutputBoundStack( stack, results, flow_of_interest );
				}
			}
		}
	}
	/**
	 * Returns the output bound for a set of <code>f_xfcaller</code> from
	 * the given <code>server</code> by recursion.<br/>
	 * This computation is nearly the same as the recursive in
	 * <code>computeRecursiveOutputBound(boolean)</code>. But the output bound
	 * is computed for the element on top of a stack. This element is a tuple
	 * with the server and the flows of interests passed as the arguments in the
	 * recursion. In many cases, the computation depends on previously computed
	 * output bounds, so when these are lacking, the calling method has to be
	 * notified. After a successful computation, the result is added to a map.
	 * 
	 * @param stack
	 *            the stack with the <code>server</code>,
	 *            <code>f_xfcaller</code>-tuple.
	 * @param results
	 *            maps the tuple to the result.
	 * @param flow_of_interest
	 *            the flow with the lowest priority
	 * @return if the output bound of the element on top of the stack could be
	 *         computed.
	 * @author Andreas Kiefer
	 */
	protected boolean decrementOutputBoundStack(	Stack<Pair<Server, Set<Flow>>> stack,
													Map<Pair<Server, Set<Flow>>, ArrivalCurve> results,
													Flow flow_of_interest )
	{
		Pair<Server, Set<Flow>> next = stack.pop();

		Server server = (Server) next.getFirst();
		Set<Flow> f_xfcaller = (Set<Flow>) next.getSecond();

		ArrivalCurve alpha_xfcaller = ArrivalCurve.NULL_ARRIVAL_CURVE;

		// This should never be true because calculations are added to the stack
		// only if isAtEnd is true. isAtEnd also checks for this condition
		if ( f_xfcaller.size() == 0 )
		{
			results.put( next, alpha_xfcaller );
			return true;
		}
			
		Set<Flow> f_server = network.getSourceFlows( server );
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
		f_xfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority
		Set<Flow> f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
		f_xxfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority

		ArrivalCurve alpha_xfcaller_server = network.getSourceFlowArrivalCurve( server, f_xfcaller_server );
		ArrivalCurve alpha_xxfcaller_server = network.getSourceFlowArrivalCurve( server, f_xxfcaller_server );
		
		for ( Link in_l : network.getInLinks( server ) )
		{
			f_server = network.getFlows( in_l );
			f_xfcaller_server = SetUtils.getIntersection( f_server, f_xfcaller );
			f_xfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority
			f_xxfcaller_server = SetUtils.getDifference( f_server, f_xfcaller_server );
			f_xxfcaller_server.remove( flow_of_interest ); // flow of interest has the lowest priority

			ArrivalCurve alpha_xfcaller_predecessor = results.get( new Pair<Server, Set<Flow>>( in_l.getSource(), f_xfcaller_server ) );
			if ( alpha_xfcaller_predecessor == null )
			{
				stack.push( next );
				return false;
			}
			ArrivalCurve alpha_xxfcaller_predecessor = results.get( new Pair<Server, Set<Flow>>( in_l.getSource(), f_xxfcaller_server ) );
			if ( alpha_xxfcaller_predecessor == null )
			{
				stack.push( next );
				return false;
			}

			alpha_xfcaller_server = ArrivalCurve.add( alpha_xfcaller_server, alpha_xfcaller_predecessor );
			alpha_xxfcaller_server = ArrivalCurve.add( alpha_xxfcaller_server, alpha_xxfcaller_predecessor );
		}

		ServiceCurve beta_loxfcaller = server.getServiceCurve();
		ServiceCurve gamma = server.getMaxServiceCurve();
		if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_ARBITRARY
			|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVERLOCAL_FLAG & server.useArbitraryMultiplexing() ) )
		{
			beta_loxfcaller = LeftOverService.arbMux( beta_loxfcaller, alpha_xxfcaller_server );
		}
		else
		{
			beta_loxfcaller = LeftOverService.fifoMux( beta_loxfcaller, alpha_xxfcaller_server );
		}

		if( Configuration.useGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useGamma() == GammaFlag.SERVERLOCAL_FLAG & server.useGamma() ) )
		{
			alpha_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alpha_xfcaller_server, gamma ), beta_loxfcaller);
		}
		else
		{
			alpha_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_server, beta_loxfcaller );
		}
		
		if( Configuration.useExtraGamma() == GammaFlag.GLOBALLY_ON
			|| ( Configuration.useExtraGamma() == GammaFlag.SERVERLOCAL_FLAG & server.useExtraGamma() ) )
		{
			alpha_xfcaller = Convolution.convolve( alpha_xfcaller, gamma.removeLatency() );
		}

		results.put( next, alpha_xfcaller );
		return true;
	}
	
	/**
	 * Recognizes if a <code>server</code> needn't be processed any further.
	 * 
	 * @param server
	 *            the server at which the output bound is computed
	 * @param f_xfcaller
	 *            set of flows for which the output bound is computed
	 * @return if the computation is at the end.
	 * 
	 * @author Andreas Kiefer
	 */
	protected boolean calculabe( Server server, Set<Flow> f_xfcaller )
	{
		if ( f_xfcaller.size() == 0 )
		{
			return true;
		}
		if ( network.inDegree( server ) == 0 )
		{
			return true;
		}
		return false;
	}
}

/**
*
* @author ngollan
*/
final class Pair<T1,T2> {
	private T1 first;
	private T2 second;

	public Pair(T1 f, T2 s) {
		first = f;
		second = s;
	}

	public T1 getFirst() {
		return first;
	}

	public T2 getSecond() {
		return second;
	}

	@Override
	public boolean equals(Object o){
		if(o == null || o.getClass() != this.getClass()) {
			return false;
		}

		@SuppressWarnings("rawtypes")
		Pair p = new Pair<Object,Object>(((Pair)o).getFirst(), ((Pair)o).getSecond());
		return (this.first != null? this.first.equals(p.first) : p.first == null)
				&& (this.second != null? this.second.equals(p.second) : p.second == null);
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 83 * hash + (this.first != null ? this.first.hashCode() : 0);
		hash = 83 * hash + (this.second != null ? this.second.hashCode() : 0);
		return hash;
	}

	@Override
	public String toString() {
		return "(" + String.valueOf(first) + ", " + String.valueOf(second) + ")";
	}
}