/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Andreas Kiefer, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;

import unikl.disco.misc.Num;
import unikl.disco.misc.SetUtils;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;

/**
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 * @author Steffen Bondorf
 */
public class PmooAnalysis extends Analysis
{
	private Set<Flow> flow_of_interest_as_set;
	
	@SuppressWarnings("unused")
	private PmooAnalysis() {}
	
	public PmooAnalysis( Network network ) {
		super.network = network;
	}

	/**
	 * Performs a pay-multiplexing-only-once (PMOO) analysis for the <code>flow_of_interest</code>.
	 * 
	 * @param flow_of_interest
	 *            the flow for which the end-to-end service curve shall be
	 *            computed
	 */
	public void performEnd2EndAnalysis( Flow flow_of_interest ) throws Exception
	{
		performAnalysisOnPath( flow_of_interest, flow_of_interest.getPath() );
	}

	public void performAnalysisOnPath( Flow flow_of_interest, Path path ) throws Exception
	{
		super.flow_of_interest = flow_of_interest;
		flow_of_interest_as_set = new HashSet<Flow>( Collections.singleton( flow_of_interest ) );
		
		beta_e2e = getPMOOServiceCurve( path, flow_of_interest_as_set, flow_of_interest );

		// single flow separated, i.e., fifo per micro flow holds
		delay_bound = DelayBound.deriveFIFO( flow_of_interest.getArrivalCurve(), beta_e2e );
		backlog_bound = Num.max( Num.ZERO, BacklogBound.derive( flow_of_interest.getArrivalCurve(), beta_e2e ) );
	}
	
	/**
	 * Returns the PMOO service curve for a set of
	 * 
	 * @param path
	 *            the link path along which service curves are concatenated (not
	 *            including the server at which the path ends)
	 * @param flows_of_interest
	 *            the set of flows to be bounded
	 * @param flow_of_interest
	 * 			  the flow of interest which will have lowest priority
	 * @return the concatenated PMOO service curve
	 * @throws Exception
	 */
	public ServiceCurve getPMOOServiceCurve( Path path, Set<Flow> flows_of_interest, Flow flow_of_interest ) throws Exception
	{
		Map<Server,Set<Flow>> map__server__joining_flows = network.getServerJoiningFlowsMap( path, flows_of_interest );
		Set<Flow> all_interfering_flows = new HashSet<Flow>();
		for ( Server s : map__server__joining_flows.keySet() ) {
			all_interfering_flows.addAll( map__server__joining_flows.get( s ) );
		}
		all_interfering_flows.removeAll( flows_of_interest ); // Should have been done by the map__server__joining_flows construction
		all_interfering_flows.remove( flow_of_interest );
		if ( all_interfering_flows.isEmpty() ) { // If there's no cross-traffic to be considered, return the convolution of a ll service curves on the path
			ServiceCurve beta_on_path = ServiceCurve.createZeroDelayBurst();
			for ( Server s : path.getServers() )
			{
				beta_on_path = Convolution.convolve( beta_on_path, s.getServiceCurve() );
			}
			return beta_on_path;
		}
		
		List<Flow> substitute_flows = new LinkedList<Flow>();
		
		// Iterate over the servers s on the path 
		for ( Server s : path.getServers() )
		{
			if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_FIFO
				|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVERLOCAL_FLAG & s.useFifoMultiplexing() ) )
			{
				throw new Exception( "PMOO analysis not available for FIFO multiplexing nodes" );
			}
			
			ArrivalCurve alpha_xf__s = ArrivalCurve.createNullArrival();

			Set<Flow> joining_flows_s = map__server__joining_flows.get( s );
			
			// Flows with the same egress server (independent of the out link) can be aggregated for PMOO's arrival bound calculation.
			// This still preserves the demultiplexing considerations.
	 		Map<Server,Set<Flow>> map__server__leaving_flows = network.getServerLeavingFlowsMap( path );

	 		// Iterate over possible egress servers of the flows joining at s and work on those sets of flows
	 		LinkedList<Server> egress_servers = path.getServers();
	 		if ( s != path.getSource() ) {
		 		egress_servers.removeAll( path.getSubPath( path.getSource(), path.getPrecedingServer( s ) ).getServers() );	 			
	 		}

	 		for ( Server egress_server : egress_servers ) {
	 			Set<Flow> xfs__s_egress = SetUtils.getIntersection( map__server__leaving_flows.get( egress_server ), joining_flows_s ); // Intersection with the remaining joining_flows prevents rejoining flows to be considered multiple times
	 			xfs__s_egress.removeAll( flows_of_interest );
	 			xfs__s_egress.remove( flow_of_interest );
	 			
	 			if ( xfs__s_egress.isEmpty() ) { // No cross-traffic to bound anymore
	 				continue;
	 			}
	 			
	 			// Get the arrival bound of flows whose source is s and egress server on the path is egress_server
	 			ArrivalCurve alpha_xf__s_egress;
				alpha_xf__s_egress = ArrivalBound.computeArrivalBound( network, s, xfs__s_egress, flow_of_interest );

		 		// Remove cross-flows flows from all local data structures to improve performance
		 		joining_flows_s.removeAll( xfs__s_egress );
		 		
		 		// Create a single substitute flow 
			 		// Name the substitute flow
			 		String substitute_flow_alias = "subst_{";
			 		for( Flow f : xfs__s_egress ) {
			 			substitute_flow_alias = substitute_flow_alias.concat( f.getAlias() + "," );
			 		}
			 		substitute_flow_alias = substitute_flow_alias.substring( 0, substitute_flow_alias.length()-1 );
			 		substitute_flow_alias = substitute_flow_alias.concat( "}" );

			 		// Get the path of the substitute flow
			 		Path substitute_flow_path = path.getSubPath( s, egress_server );
				 		
		 		Flow subst_xf__s_egress = Flow.createDummyFlow( substitute_flow_alias, alpha_xf__s_egress, substitute_flow_path );
		 		substitute_flows.add( subst_xf__s_egress );
			 		
		 		// Add the arrival bound for the substitute flow to the arrival bound of server s
				alpha_xf__s = ArrivalCurve.add( alpha_xf__s, alpha_xf__s_egress );
			}
	 		map__server__alpha.put( s, alpha_xf__s );
		}
		
		return computePMOOServiceCurve( path, substitute_flows );
	}

	/**
	 * Concatenates the service curves along the given path <code>path</code>
	 * according to the PMOO approach and returns the result.<br/>
	 * It first decomposes all arrival curves (service curves) into token
	 * buckets (rate latency curves), enumerates over all combinations of token
	 * buckets and rate latency curves, and calls
	 * <code>computePartialPMOOServiceCurve()</code> for each combination. The
	 * total PMOO service curve is the maximum of all partial service curves.
	 * 
	 * @param path
	 *            link path traversing the servers whose service curves shall be
	 *            concatenated (excluding the one at which the path ends).
	 * @param cross_flow_substitutes
	 *            map of flows to egress points (integer index into the path)
	 * @return the PMOO service curve
	 */
	protected static ServiceCurve computePMOOServiceCurve( Path path, List<Flow> cross_flow_substitutes ) {		
		// create a flow-->tb_iter map
		Map<Flow, Integer> flow_tb_iter_map = new HashMap<Flow, Integer>();
		for ( Flow f : cross_flow_substitutes )
		{
			flow_tb_iter_map.put( f, Integer.valueOf( 0 ) );
		}
		// create a list of rl_iters
		int number_servers = path.getServers().size();
		ServiceCurve[] service_curves = new ServiceCurve[number_servers];
		int[] server_rl_iters = new int[number_servers];
		int[] server_rl_counts = new int[number_servers];
		int i = 0;
		for ( Server server : path.getServers() )
		{
			ServiceCurve service_curve = server.getServiceCurve();
			service_curves[i] = service_curve;
			server_rl_iters[i] = 0;
			server_rl_counts[i] = service_curve.getRLComponentCount();
			i++;
		}

		ServiceCurve beta_total = ServiceCurve.NULL_SERVICE_CURVE;

		boolean more_combinations = true;
		while ( more_combinations )
		{
			// compute service curve for this combination
			ServiceCurve beta = computePartialPMOOServiceCurve(	path,
															service_curves,
															cross_flow_substitutes,
															flow_tb_iter_map,
															server_rl_iters );
			if ( beta != ServiceCurve.NULL_SERVICE_CURVE )
			{
				beta_total = ServiceCurve.max( beta_total, beta );
			}

			// first check whether there are more combinations of flow TBs
			more_combinations = false;
			for ( Flow f : flow_tb_iter_map.keySet() )
			{
				ArrivalCurve f_bound = f.getArrivalCurve();

				i = flow_tb_iter_map.get( f ).intValue();
				if ( i + 1 < f_bound.getTBComponentCount() )
				{
					flow_tb_iter_map.put( f, Integer.valueOf( i + 1 ) );
					more_combinations = true;
					break;
				}
				else
				{
					flow_tb_iter_map.put( f, Integer.valueOf( 0 ) );
				}
			}

			// if not, check whether there are more combinations of server RLs
			if ( !more_combinations )
			{
				for ( i = 0; i < server_rl_iters.length; i++ )
				{
					int j = server_rl_iters[i];
					if ( j + 1 < server_rl_counts[i] )
					{
						server_rl_iters[i] = j;
						more_combinations = true;
						break;
					}
					else
					{
						server_rl_iters[i] = 0;
					}
				}
			}
		}

		return beta_total;
	}

	/**
	 * Calculates the partial PMOO service curve for the given flow set by
	 * combining all servers having an outgoing link contained in the given
	 * link-path. For each flow considers only one of its token bucket
	 * components (selected via the flow_tb_iter_map) and for each service curve
	 * considers only one rate latency curve (selected via the server_rl_iters).
	 * 
	 * @return a partial PMOO service curve
	 */
	protected static ServiceCurve computePartialPMOOServiceCurve(	Path path,
			ServiceCurve[] service_curves,
															List<Flow> cross_flow_substitutes,
															Map<Flow, Integer> flow_tb_iter_map,
															int[] server_rl_iters )
	{
		Num T = Num.ZERO;
		Num R = Num.POSITIVE_INFINITY;
		Num sum_bursts = Num.ZERO;
		Num sum_latencyterms = Num.ZERO;
		
		Set<Flow> present_flows = new HashSet<Flow>();
		for ( Server s : path.getServers() )
		{
			int i = path.getServers().indexOf( s );
			// add incoming flows
			for ( Flow f : cross_flow_substitutes ) {
				if ( f.getPath().getServers().contains( s ) ) {
					present_flows.add( f );
				}
			}

			ServiceCurve current_rl = service_curves[i].getRLComponent( server_rl_iters[i] );

			// sum up latencies
			T = Num.add( T, current_rl.getLatency() );

			// calculate and store sum of rates of all passing flows
			Num sum_r = Num.ZERO;
			for ( Flow f : present_flows )
			{
				ArrivalCurve bound = f.getArrivalCurve();
				ArrivalCurve current_tb = bound.getTBComponent( ((Integer) flow_tb_iter_map.get( f )).intValue() );
				sum_r = Num.add( sum_r, current_tb.getSustainedRate() );
			}

			// update latency terms (increments)
			sum_latencyterms = Num.add( sum_latencyterms, Num.mult( sum_r, current_rl.getLatency() ) );

			// compute residual rate; update min
			Num Ri = Num.sub( current_rl.getSustainedRate(), sum_r );
			if ( Ri.le( Num.ZERO ) )
			{
				return ServiceCurve.createNullService();
			}
			R = Num.min( R, Ri );

			// remove all outgoing flows from the set of present flows
			Set<Flow> leaving_flows = new HashSet<Flow>();
			for ( Flow f : present_flows ) {
				if ( path.getServers().indexOf( f.getSink() ) <= i ) {
					leaving_flows.add( f );
				}
			}
			present_flows.removeAll( leaving_flows );
		}

		// compute sum of bursts
		for ( Flow f : cross_flow_substitutes )
		{
			ArrivalCurve bound = f.getArrivalCurve();
			ArrivalCurve current_tb = bound.getTBComponent( ((Integer) flow_tb_iter_map.get( f )).intValue() );
			sum_bursts = Num.add( sum_bursts, current_tb.getTBBurst() );
		}

 		T = Num.add( T,Num.div( Num.add( sum_bursts, sum_latencyterms ), R ) );

 		if( T == Num.POSITIVE_INFINITY ) {
 			return ServiceCurve.createNullService();
 		}
 		if( R == Num.POSITIVE_INFINITY ) {
 			return ServiceCurve.createBurstDelay( T );
 		}

		return ServiceCurve.createRateLatency( R, T );
	}
}
