/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Set;
import java.util.HashSet;

import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.Configuration.GammaFlag;
import unikl.disco.network.Flow;
import unikl.disco.network.Link;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;
import unikl.disco.minplus.Deconvolution;

public class PmooArrivalBound extends ArrivalBound {
	
	@SuppressWarnings("unused")
	private PmooArrivalBound() {}
	
	public PmooArrivalBound( Network network ) {
		super.network = network;
	}

	protected ArrivalCurve computeArrivalBound( Server server ) throws Exception {
		return computeArrivalBound( server, network.getFlows( server ), Flow.NULL_FLOW );
	}

	protected ArrivalCurve computeArrivalBound( Server server, Set<Flow> f_xfcaller ) throws Exception {
		return computeArrivalBound( server, f_xfcaller, Flow.NULL_FLOW );
	}
	
	protected ArrivalCurve computeArrivalBound( Link link ) throws Exception {
		return computeArrivalBound( link.getDest(), network.getFlows( link ), Flow.NULL_FLOW );
	}
	
	protected ArrivalCurve computeArrivalBound( Link link, Set<Flow> f_xfcaller ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( f_xfcaller, network.getFlows( link ) );
		return computeArrivalBound( link.getDest(), flows_on_l_to_bound, Flow.NULL_FLOW );
	}

	protected ArrivalCurve computeArrivalBound( Link link, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws Exception {
		Set<Flow> flows_on_l_to_bound = SetUtils.getIntersection( f_xfcaller, network.getFlows( link ) );
		return computeArrivalBound( link.getDest(), flows_on_l_to_bound, flow_of_interest );
	}

	/**
	 * Computes the PMOO arrival bound for a set of
	 * <code>flows_to_bound</code>. The difference to the standard output
	 * bound method is that this method tries to compute tighter bounds by
	 * concatenating as many servers as possible using the PMOO approach. It
	 * does so by searching from <code>server</code> towards the sinks of the
	 * flows contained in <code>flows_of_interest</code> until it reaches the
	 * server where all these flows first meet each other (the
	 * "splitting point"). It then concatenates all servers between the
	 * splitting point (inclusive) and <code>server</code> (exclusive).<br/>
	 * Note: The previous sentence means that the output bound is for the flows
	 * <b>before</b> <code>server</code> and not after the server as is the case
	 * with the standard computateOutputBound().
	 * 
	 * @param server
	 *            that all flows of interest flow into
	 * @param f_xfcaller
	 *            the set of flows of interest
	 * @return the PMOO output bound
	 * @throws Exception 
	 */
	protected ArrivalCurve computeArrivalBound( Server server, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws Exception
	{
		ArrivalCurve arrival_bound = ArrivalCurve.createNullArrival();
		
		// Get flows of interest
		Set<Flow> f_xfcaller_server = SetUtils.getIntersection( f_xfcaller, network.getFlows( server ) );
		f_xfcaller_server.remove( flow_of_interest );
		if ( f_xfcaller_server.size() == 0 )
		{
			return arrival_bound;
		}

		// Get cross-traffic originating in server
		Set<Flow> f_xfcaller_sourceflows_server = SetUtils.getIntersection( f_xfcaller_server, network.getSourceFlows( server ) );
		arrival_bound = network.getSourceFlowArrivalCurve( server, f_xfcaller_sourceflows_server );

		// Get cross-traffic from each predecessor. Call per link in order to get splitting points
		ArrivalCurve arrival_bound_predec = ArrivalCurve.NULL_ARRIVAL_CURVE;
		Set<Flow> f_xfcaller_server_bounded = f_xfcaller_sourceflows_server;
		if ( !f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
			for ( Link in_l : network.getInLinks( server ) ) {
				Set<Flow> f_xfcaller_predec = SetUtils.getIntersection( network.getFlows( in_l ), f_xfcaller_server );
				
				if( f_xfcaller_predec.isEmpty() ) { // Do not check links without flows of interest
					continue;
				}
	
				arrival_bound_predec = computeArrivalBoundOnLink( in_l, f_xfcaller_predec, flow_of_interest );
				arrival_bound = ArrivalCurve.add( arrival_bound, arrival_bound_predec );
				
				f_xfcaller_server_bounded.addAll( f_xfcaller_predec ); // Stop as soon as all flows are bounded
				if ( f_xfcaller_server_bounded.containsAll( f_xfcaller_server ) ) {
					break;
				}
			}
		}

		return arrival_bound;
	}
	
	private ArrivalCurve computeArrivalBoundOnLink( Link link, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws Exception {
		ArrivalCurve alpha_xfcaller = ArrivalCurve.createNullArrival();
		if ( f_xfcaller.size() == 0 )
		{
			return alpha_xfcaller;
		}
		
		// Get the common sub-path of f_xfcaller flows crossing the given link
		Server to = link.getDest();
		Set<Flow> f_to = network.getFlows( to );
		Set<Flow> f_xfcaller_to = SetUtils.getIntersection( f_to, f_xfcaller );
		f_xfcaller_to.remove( flow_of_interest );
		if ( f_xfcaller_to.size() == 0 )
		{
			return alpha_xfcaller;
		}
		Server from = network.findSplittingServer( to, f_xfcaller_to );
		Flow f_representative = f_xfcaller_to.iterator().next();
		
		Path common_subpath = f_representative.getSubPath( from, f_representative.getPath().getPrecedingServer( to ) );

		// Get left-over service curve for the sub-path
		PmooAnalysis pmoo = new PmooAnalysis( network );
		ServiceCurve beta_loxfcaller_subpath = pmoo.getPMOOServiceCurve( common_subpath, f_xfcaller, flow_of_interest );
		
		// Get arrival bound at the splitting point:
		// We need to know the arrival bound of f_xfcaller at the server 'from', i.e., at the above sub-path's source
		// in order to deconvolve it with beta_loxfcaller_subpath to get the arrival bound of the sub-path
		// Note that flows f_xfcaller that originate in 'from' are covered by this call of computeArrivalBound
		ArrivalCurve alpha_xfcaller_from = computeArrivalBound( from, f_xfcaller, flow_of_interest );
		
		// Convolve to get the bound
		Set<Server> servers_on_subpath = new HashSet<Server>( common_subpath.getServers() );
		// see "Improving Performance Bounds in Feed-Forward Networks by Paying Multiplexing Only Once", Lemma 2
		if( Configuration.useGamma() != GammaFlag.GLOBALLY_OFF  )
		{
			ServiceCurve gamma = Network.getGamma( servers_on_subpath );
			alpha_xfcaller = Deconvolution.deconvolve( Convolution.convolve( alpha_xfcaller_from, gamma), beta_loxfcaller_subpath );
		}
		else
		{
			alpha_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_from, beta_loxfcaller_subpath );
		}
		
		if( Configuration.useExtraGamma() != GammaFlag.GLOBALLY_OFF )
		{
			ServiceCurve extra_gamma = Network.getExtraGamma( servers_on_subpath );
			alpha_xfcaller = Convolution.convolve( alpha_xfcaller, extra_gamma );
		}
		
		return alpha_xfcaller;
	}
}
