/*
 *  (c) 2008 - 2014 Andreas Kiefer, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.nc;

import java.util.Map;
import java.util.Set;
import java.util.HashMap;

import unikl.disco.misc.Num;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.nc.Configuration.MuxDiscipline;
import unikl.disco.network.Flow;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.minplus.Convolution;

/**
 * @author Frank A. Zdarsky
 * @author Andreas Kiefer
 * @author Steffen Bondorf
 */
public class SeparateFlowAnalysis extends Analysis
{
	Map<Server, ServiceCurve> map__server__beta_lo = new HashMap<Server, ServiceCurve>();

	@SuppressWarnings("unused")
	private SeparateFlowAnalysis(){}
	
	public SeparateFlowAnalysis( Network network ) {
		super.network = network;
	}
	/**
	 * Performs a separated flow analysis for the <code>flow_of_interest</code>
	 * and returns the end-to-end service curve.<br/>
	 * This analysis first blends out the flow of interest and then computes for
	 * each server along this flow's path the left-over service curve that
	 * results if all remaining flows crossing this server receive their maximum
	 * amount of service. Then all left-over service curves are concatenated to
	 * receive the end-to-end service curve from the perspective of the flow of
	 * interest.
	 * 
	 * @param flow_of_interest
	 *            the flow for which the end-to-end service curve shall be
	 *            computed
	 */
	public void performEnd2EndAnalysis( Flow flow_of_interest ) throws Exception
	{
		performAnalysisOnPath( flow_of_interest, flow_of_interest.getPath() );
	}

	public void performAnalysisOnPath( Flow flow_of_interest, Path path ) throws Exception
	{
		super.flow_of_interest = flow_of_interest;
		map__server__beta_lo = new HashMap<Server, ServiceCurve>();
		beta_e2e = ServiceCurve.createZeroDelayBurst();
		
		for ( Server server : path.getServers() )
		{
			ServiceCurve beta_lofoi = getSfaServiceCurve( network, server, flow_of_interest );

			map__server__beta_lo.put( server, beta_lofoi );
			
			// Convolve all left over service curves, server by server
			beta_e2e = Convolution.convolve( beta_e2e, beta_lofoi );
		}
		
		// single flow separated, i.e., fifo per micro flow holds
		delay_bound = DelayBound.deriveFIFO( flow_of_interest.getArrivalCurve(), beta_e2e );
		backlog_bound = Num.max( Num.ZERO, BacklogBound.derive( flow_of_interest.getArrivalCurve(), beta_e2e ) );
	}
	
	public ServiceCurve getSfaServiceCurve( Network network, Server server, Flow flow_of_interest ) throws Exception {
		Set<Flow> f_xfoi_server = network.getFlows( server );		
		f_xfoi_server.remove( flow_of_interest );

		ServiceCurve beta_lofoi = server.getServiceCurve();
		
		if ( f_xfoi_server.isEmpty() ) {
			return beta_lofoi;
		}
															// network, server, flows_to_bound, flows_lower_priority
		ArrivalCurve alpha_xfoi = ArrivalBound.computeArrivalBound( network, server, f_xfoi_server, flow_of_interest );
		map__server__alpha.put( server, alpha_xfoi );
		
		// Calculate the left-over service curve for the flow of interest
		if( Configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_FIFO
			|| ( Configuration.multiplexingDiscipline() == MuxDiscipline.SERVERLOCAL_FLAG & server.useFifoMultiplexing() ) )
		{
			beta_lofoi = LeftOverService.fifoMux( beta_lofoi, alpha_xfoi );
		}
		else
		{
			beta_lofoi = LeftOverService.arbMux( beta_lofoi, alpha_xfoi );				
		}
		return beta_lofoi;
	}

	public Map<Server, ServiceCurve> getServerLeftOverBetaMap(){
		return new HashMap<Server, ServiceCurve>( map__server__beta_lo );
	}
}
