/*
 *  (c) 2005 - 2014 Frank A. Zdarsky, Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.network;

import java.util.LinkedList;
import java.util.List;

import unikl.disco.curves.ArrivalCurve;
//import unikl.disco.curves.Curve;

/**
 * Class representing flows through the network.
 *
 * @author Frank A. Zdarsky
 * @author Steffen Bondorf
 */
public class Flow {
	public static final Flow NULL_FLOW = createDummyFlow( "null", ArrivalCurve.NULL_ARRIVAL_CURVE, Path.createNullPath() ); 
	
	/** The flow's ID. */
	private int	id;
	/** The flow's arrival curve */
	private ArrivalCurve arrival_curve;
	
	private String alias;
	/** The link path the flow traverses incl explicit sink */
	private Path path;
	
	/**
	 * Creates a dummy flow with an arrival curve.
	 * @return a dummy flow
	 */
	public static Flow createDummyFlow( String alias, ArrivalCurve ac, Path path ) {
		Flow result = new Flow();
		result.alias = alias;
		result.arrival_curve = ac;
		result.path = path;
		return result;
	}
	
	/**
	 * @param source the server at which the flow originates
	 * @param sink the server at which the flow ends
	 * @param ac the flow's arrival curve
	 * @param path the link path the flow traverses
	 */
	protected Flow( int id, String alias, ArrivalCurve ac, Path path ) throws Exception {
		this.id = id;
		this.alias = alias;
		this.arrival_curve = ac;
		this.path   = path;
	}
	
	private Flow() {
		this.id     = -1;
		this.arrival_curve     = null;
		this.path   = null;
	}
	
	public boolean setArrivalCurve( ArrivalCurve arrival_curve ) {
		this.arrival_curve = arrival_curve;
		return true;
	}

	/**
	 * 
	 * @return a copy of the arrival curve
	 */
	public ArrivalCurve getArrivalCurve() {
		return arrival_curve.copy();
	}
	
	public int getId() {
		return id;
	}
	
	public String getAlias() {
		return alias;
	}
	
	public void setAlias( String alias ) {
		this.alias = alias;
	}

	public Path getPath() {
		return path;
	}

	/**
	 * Returns a string representation of the flow.
	 * @return a string representation of the flow
	 */
	@Override
	public String toString() {
		return "(" + id + "," + alias.toLowerCase() + "," + path.toString() + "," + arrival_curve.toString() + ")";
	}

	// --------------------------------------------------------------------------------------------
	// Shortcuts to conveniently access the path's according methods  
	// --------------------------------------------------------------------------------------------
	public Server getSource() {
		return path.getSource();
	}
	
	public Server getSink() {
		return path.getSink();
	}

    public LinkedList<Server> getServersOnPath() throws Exception {
    	return new LinkedList<Server>( path.getServers() );
    }

    public List<Link> getLinksOnPath() throws Exception {
    	return new LinkedList<Link>( path.getLinks() );
    }
	
    /**
     * @param from inclusive
     * @param to inclusive
     */
    public Path getSubPath( Server from, Server to ) throws Exception {
    	return path.getSubPath( from, to );
    }
    
	public Link getPrecedingLink( Server s ) throws Exception {
		return path.getPrecedingLink( s );
	}
	
 	public Link getSucceedingLink( Server s ) throws Exception {
 		return path.getSucceedingLink( s );
 	}

 	public Server getPrecedingServer( Server s ) throws Exception {
 		return path.getPrecedingServer( s );
 	}

 	public Server getSucceedingServer( Server s ) throws Exception {
 		return path.getSucceedingServer( s );
 	}
}