/*
 *  (c) 2011 - 2014 Steffen Bondorf,
 *  
 *  				disco | Distributed Computer Systems Lab
 *                  University of Kaiserslautern, Germany
 *                                  
 *         All Rights Reserved.
 *
 *  This software is work in progress and is released in the hope that it will
 *  be useful to the scientific community. It is provided "as is" without
 *  express or implied warranty, including but not limited to the correctness
 *  of the code or its suitability for any particular purpose.
 *
 *  You are free to use this software for any non-commercial educational or
 *  research purpose, provided that this copyright notice is not removed or
 *  modified. For commercial uses please contact the respective author(s).
 *
 *  If you find our software useful, we would appreciate if you mentioned it
 *  in any publication arising from the use of this software or acknowledge
 *  our work otherwise. We would also like to hear of any fixes or useful
 *  extensions to this software.
 *
 */

package unikl.disco.network;

import java.util.Collection;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Collections;

import unikl.disco.curves.ArrivalCurve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.minplus.Convolution;

/**
 * 
 * @author Steffen Bondorf
 *
 */
public class Network {
	private Set<Server> servers;
	private Set<Link> links;
	private Set<Flow> flows;
	
	private Map<Server,Set<Link>> map__server__in_links;
	private Map<Server,Set<Link>> map__server__out_links;
	
	private Map<Server,Set<Flow>> map__server__flows;
	private Map<Server,Set<Flow>> map__server__source_flows;

	private Map<Link,Set<Flow>> map__link__flows;
	
	private String server_default_name_prefix = "s";
	private int server_id_counter = 0;
	protected Map<Integer,Server> map__id__server;

	private String link_default_name_prefix = "l";
	private int link_id_counter = 0;

	private String flow_default_name_prefix = "f";
	private int flow_id_counter = 0;
	
	public Network() {
		servers = new HashSet<Server>();
		links = new HashSet<Link>();
		flows = new HashSet<Flow>();
		
		map__id__server = new HashMap<Integer,Server>();
		
		map__server__in_links = new HashMap<Server,Set<Link>>();
		map__server__out_links = new HashMap<Server,Set<Link>>();

		map__server__flows = new HashMap<Server,Set<Flow>>();
		map__server__source_flows = new HashMap<Server,Set<Flow>>();
		
		map__link__flows = new HashMap<Link,Set<Flow>>();
	}

	// --------------------------------------------------------------------------------------------
	// servers
	// --------------------------------------------------------------------------------------------
	public Server addServer( ServiceCurve service_curve ) {
		return addServer( service_curve.copy(), ServiceCurve.createZeroDelayBurst(), true, false, false );
	}
	
	public Server addServer( ServiceCurve service_curve, boolean arbitrary_mux ) {
		return addServer( service_curve, ServiceCurve.createZeroDelayBurst(), arbitrary_mux, true, true );
	}
	
	public Server addServer( String alias, ServiceCurve service_curve ) {
		return addServer( alias, service_curve.copy(), ServiceCurve.createZeroDelayBurst(), true, false, false );
	}
	
	public Server addServer( String alias, ServiceCurve service_curve, boolean arbitrary_mux ) {
		return addServer( alias, service_curve.copy(), ServiceCurve.createZeroDelayBurst(), arbitrary_mux, false, false );
	}
	
	/**
	 * By default the server's use_gamma and use_extra_gamma are enabled
	 * 
	 * @param service_curve
	 * @param max_service_curve
	 * @return the added server
	 */
	public Server addServer( ServiceCurve service_curve, ServiceCurve max_service_curve ) {
		return addServer( service_curve, max_service_curve, true, true, true );
	}
	
	/**
	 * By default the server's use_gamma and use_extra_gamma are enabled
	 * 
	 * @param alias
	 * @param service_curve
	 * @param max_service_curve
	 * @return the added server
	 */
	public Server addServer( String alias, ServiceCurve service_curve, ServiceCurve max_service_curve ) {
		return addServer( alias , service_curve, max_service_curve, true, true, true );
	}

	public Server addServer( ServiceCurve service_curve, ServiceCurve max_service_curve, boolean use_gamma, boolean use_extra_gamma ) {
		return addServer( service_curve, max_service_curve, true, use_gamma, use_extra_gamma );
	}
	
	public Server addServer( String alias, ServiceCurve service_curve, ServiceCurve max_service_curve, boolean arbitrary_mux ) {
		return addServer( alias , service_curve, max_service_curve, arbitrary_mux, true, true );		
	}
	
	public Server addServer( ServiceCurve service_curve, ServiceCurve max_service_curve, boolean arbitrary_mux ) {
		return addServer( service_curve.copy(), max_service_curve.copy(), arbitrary_mux, true, true );
	}
	
	public Server addServer( String alias, ServiceCurve service_curve, ServiceCurve max_service_curve, boolean use_gamma, boolean use_extra_gamma ) {
		return addServer( alias, service_curve.copy(), max_service_curve.copy(), true, use_gamma, use_extra_gamma );
	}
	
	public Server addServer( ServiceCurve service_curve, ServiceCurve max_service_curve, boolean arbitrary_mux, boolean use_gamma, boolean use_extra_gamma ) {
		String alias = server_default_name_prefix + Integer.toString( server_id_counter );
		return addServer( alias, service_curve, max_service_curve, arbitrary_mux, use_gamma, use_extra_gamma );
	}
	
	public Server addServer( String alias, ServiceCurve service_curve, ServiceCurve max_service_curve, boolean arbitrary_mux, boolean use_gamma, boolean use_extra_gamma ) {
		Server new_server = new Server( server_id_counter, alias, service_curve.copy(), max_service_curve.copy(), arbitrary_mux, use_gamma, use_extra_gamma );
		
		Integer integer_object = new Integer( server_id_counter );
		map__id__server.put( integer_object, new_server );
		
		server_id_counter++;
		
		map__server__in_links.put( new_server, new HashSet<Link>() );
		map__server__out_links.put( new_server, new HashSet<Link>() );
		
		map__server__flows.put( new_server, new HashSet<Flow>() );
		map__server__source_flows.put( new_server, new HashSet<Flow>() );
		
		servers.add( new_server );
		
		map__id__server.put( integer_object, new_server );
		
		return new_server;
	}
	
	public void removeServer( Server s ) throws Exception {
		if ( !servers.contains( s ) ) {
			throw new Exception( "Server to be removed is not in this network's list of servers" );
		}
		
		map__id__server.remove( s.getId() );
		
		// Remove all flows that cross this server
		for( Flow f : map__server__flows.get( s ) ) {
			removeFlow( f );
		}
		map__server__flows.remove( s );
		
		// Remove all links that are connected to the server
		for( Link l : getIncidentLinks( s ) ) {
			removeLink( l );
		}
		map__server__in_links.remove( s );
		map__server__out_links.remove( s );
		map__server__source_flows.remove( s );

		// Remove the server
		servers.remove( s );
	}
	
	public Set<Flow> getSourceFlows( Server source ) {
		return new HashSet<Flow>( map__server__source_flows.get( source ) );
	}
	
	public Server getServer( int id ) throws Exception {
		if ( id < 0 || id > map__id__server.size() - 1 ) {
			throw new Exception( "No server with id " + Integer.toString(id) + " found" );
		}

		Server server = map__id__server.get( new Integer( id ) );
		
		if ( server == null ) {
			throw new Exception( "No server with id " + Integer.toString(id) + " found" );
		}
		
		return server;
	}
	
	public Set<Server> getServers() {
		return new HashSet<Server>( servers );
	}
	
	public int numServers() {
		return servers.size();
	}

	public int degree( Server s ) {
		return inDegree( s ) + outDegree( s );
	}

	public int inDegree( Server s ) {
		return getInLinks( s ).size();
	}

	public int outDegree( Server s ) {
		return getOutLinks( s ).size();
	}

	/**
	 * Returns a new set consisting of references to the servers.
	 * 
	 * @param s
	 * @return the incoming links of s
	 */
	public Set<Link> getInLinks( Server s ) {
		return new HashSet<Link>( map__server__in_links.get( s ) );
	}
	
	/**
	 * Returns a new set consisting of references to the servers.
	 * 
	 * @param s
	 * @return the outgoing links of s
	 */
	public Set<Link> getOutLinks( Server s ) {
		return new HashSet<Link>( map__server__out_links.get( s ) );
	}
	
	/**
	 * Returns a new set consisting of references to the links.
	 * 
	 * @param s
	 * @return the incident links
	 */
	public Set<Link> getIncidentLinks( Server s ) {
		return SetUtils.getUnion( getInLinks( s ), getOutLinks( s ) );
	}
	
	/**
	 * Returns a new set consisting of references to the servers.
	 * 
	 * @param s
	 * @return the neighboring servers of s
	 */
	public Set<Server> getNeighbors( Server s ) {
		return SetUtils.getUnion( getSuccessors( s ), getPredecessors( s ) );
	}
	
	/**
	 * Returns a new set consisting of references to the servers.
	 * 
	 * @param s
	 * @return the source servers of incoming links of s
	 */
	public Set<Server> getPredecessors( Server s ) {		
		Set<Server> predecessors = new HashSet<Server>();
		for( Link l : getInLinks( s ) ) {
			predecessors.add( l.getSource() );
		}
		return predecessors;
	}
	
	/**
	 * Returns a new set consisting of references to the servers.
	 * 
	 * @param s
	 * @return the sink servers of outgoing links of s
	 */
	public Set<Server> getSuccessors( Server s ) {
		Set<Server> successors = new HashSet<Server>();
		for( Link l : getOutLinks( s ) ) {
			successors.add( l.getDest() );
		}
		return successors;
	}

	// --------------------------------------------------------------------------------------------
	// links
	// --------------------------------------------------------------------------------------------
	public Link addLink( Server source, Server destination ) throws Exception {
		String alias = link_default_name_prefix + Integer.toString( link_id_counter );
		return addLink( alias, source, destination ); 
	}

	public Link addLink( String alias, Server source, Server destination ) throws Exception {
		if( !servers.contains( source ) ){
			throw new Exception( "link's source not present in the network" );
		}
		if ( !servers.contains( destination ) ) {
			throw new Exception( "link's destination not present in the network" );
		}
		
		try {
			// this implicitly signals the caller that the link was already present in the network 
			// by returning a link with a name different to the given one
			return findLink( source, destination );
		} catch (Exception e) {
			Link new_link = new Link( link_id_counter, alias, source, destination );
			link_id_counter++;
			
			map__link__flows.put( new_link, new HashSet<Flow>() );
			
			map__server__in_links.get( destination ).add( new_link );
			map__server__out_links.get( source ).add( new_link );
			
			links.add( new_link );
			return new_link;
		}
	}

	public void removeLink( Link l ) throws Exception {
		if( !links.contains( l ) ){
			throw new Exception( "Link to be removed is not in this network's list of links" );
		}

		// Remove all flows on the link
		for( Flow f : map__link__flows.get( l ) ) {
			removeFlow( f );
		}
		map__link__flows.remove( l );
		map__server__in_links.get( l.getDest() ).remove( l );
		map__server__out_links.get( l.getSource() ).remove( l );
		
		links.remove( l );
	}

	public Set<Link> getLinks() {
		return new HashSet<Link>( links );
	}

	/**
	 * 
	 * @param src
	 * @param dest
	 * @return the link from src to dest 
	 */
	public Link findLink( Server src, Server dest ) throws Exception {
		Set<Link> connecting_link_as_set = SetUtils.getIntersection( getInLinks( dest ), getOutLinks( src ) );
		if ( connecting_link_as_set.isEmpty() ) {
			throw new Exception ( "No link between " + src.toString() + " and " + dest.toString() + " found." );
		}else{
			if ( connecting_link_as_set.size() > 1 ) {
				throw new Exception ( "Too many links between " + src.toString() + " and " + dest.toString() + " found." );
			} else {
				return connecting_link_as_set.iterator().next();
			}
		}
	}
	
	// --------------------------------------------------------------------------------------------
	// flows
	//
	// INFO It is not allowed to change the path of a flow,
	// 		i.e., the user needs to remove the old flow and insert a new one with the modified path
	// --------------------------------------------------------------------------------------------
	public Flow addFlow( ArrivalCurve arrival_curve, List<Link> path ) throws Exception {
		String alias = flow_default_name_prefix + Integer.toString( flow_id_counter );
		return addFlow( alias, arrival_curve, path );
	}
	
	/**
	 * Creates a flow and adds it to the network.
	 * 
	 * @param arrival_curve
	 * @param source
	 * @param sink
	 * @return the flow
	 * @throws Exception
	 */
	public Flow addFlow( ArrivalCurve arrival_curve, Server source, Server sink ) throws Exception {
		String alias = flow_default_name_prefix + Integer.toString( flow_id_counter );
		return addFlow( alias, arrival_curve, source, sink );
	}
	
	/**
	 * Creates a flow and adds it to the network.
	 * 
	 * @param alias
	 * @param arrival_curve
	 * @param source
	 * @param sink
	 * @return the flow
	 * @throws Exception
	 */
	public Flow addFlow( String alias, ArrivalCurve arrival_curve, Server source, Server sink ) throws Exception {
		if( source == sink ) {
			return addFlow( arrival_curve, sink );
		}
		
		return addFlowToNetwork( alias, arrival_curve, getShortestPath( source, sink ) );
	}

	/**
	 * Creates a flow and adds it to the network.
	 * 
	 * @param alias
	 * @param arrival_curve
	 * @param path
	 * @return the flow
	 * @throws Exception
	 */
	public Flow addFlow( String alias, ArrivalCurve arrival_curve, List<Link> path ) throws Exception {
		return addFlowToNetwork( alias, arrival_curve, createPath( path ) );
	}

	/**
	 * Creates a flow and adds it to the network.
	 * 
	 * @param arrival_curve
	 * @param single_hop
	 * @return the flow
	 * @throws Exception
	 */
	public Flow addFlow( ArrivalCurve arrival_curve, Server single_hop ) throws Exception {		
		String alias = flow_default_name_prefix + Integer.toString( flow_id_counter );
		return addFlow( alias, arrival_curve, single_hop );
	}
	
	/**
	 * Creates a flow and adds it to the network.
	 * 
	 * @param alias
	 * @param arrival_curve
	 * @param single_hop
	 * @return the flow
	 * @throws Exception
	 */
	public Flow addFlow( String alias, ArrivalCurve arrival_curve, Server single_hop ) throws Exception {
		return addFlowToNetwork( alias, arrival_curve, createPath( single_hop ) );
	}
	
	// needed to be named differently due to a collision of the method's signature with a user visible one's
	private Flow addFlowToNetwork( String alias, ArrivalCurve arrival_curve, Path path ) throws Exception {
		Flow new_flow = new Flow( flow_id_counter, alias, arrival_curve.copy(), path );
		flow_id_counter++;
		
		flows.add( new_flow );
		map__server__source_flows.get( path.getSource() ).add( new_flow );
		
		for ( Link l : path.getLinks() )
		{
			map__link__flows.get( l ).add( new_flow );
		}
		for ( Server s : path.getServers() )
		{
			map__server__flows.get( s ).add( new_flow );
		}
		
		return new_flow;
	}
	
	/**
	 * Removes a flow from the network.
	 * 
	 * @param flow the flow to be removed
	 */
	public void removeFlow( Flow flow ) throws Exception {
		if ( !flows.contains( flow ) ) {
			throw new Exception( "Flow to be removed is not in this network's list of flows" );
		}
		
		for ( Link l : flow.getPath().getLinks() )
		{
			map__link__flows.get( l ).remove( flow );
		}
		for ( Server s : flow.getPath().getServers() )
		{
			map__server__flows.get( s ).remove( flow );
		}

		map__server__source_flows.get( flow.getSource() ).remove( flow );
		flows.remove( flow );
	}
	
	public Set<Flow> getFlows() {
		return new HashSet<Flow>( flows );
	}
	
	public Set<Flow> getFlows( Link l ) {
		return new HashSet<Flow>( map__link__flows.get( l ) );
	}

	public Set<Flow> getFlows( Set<Link> links ) {
		HashSet<Flow> flows = new HashSet<Flow>();
		
		for ( Link l : links ) {
			flows.addAll( map__link__flows.get( l ) );
		}
		
		return flows;
	}
	
	public Set<Flow> getFlows( Server s )
	{
		return new HashSet<Flow>( map__server__flows.get( s ) );
	}

	/**
	 * Returns an aggregate arrival curve for all flows originating in <code>source</code>.
	 * 
	 * Returns a null curve if there are no source flows at <code>source</code>.
	 * 
	 * @param source
	 *            the source of all flows to be aggregated
	 * @return an aggregate arrival curve
	 */
	public ArrivalCurve getSourceFlowArrivalCurve( Server source )
	{
		return getSourceFlowArrivalCurve( source, getSourceFlows( source ) );
	}
	
	/**
	 * Returns an aggregate arrival curve for all flows originating in
	 * <code>source</code> and contained in the set <code>outgoing_flows</code>.
	 * 
	 * Returns a null curve if the intersection of those sets of curves is empty.
	 * 
	 * @param source
	 *            the source of all flows to be aggregated
	 * @param outgoing_flows
	 *            the set of all flows to be aggregated
	 * @return an aggregate arrival curve
	 */
	public ArrivalCurve getSourceFlowArrivalCurve( Server source, Set<Flow> outgoing_flows )
	{
		ArrivalCurve a_out = ArrivalCurve.createNullArrival();
		
		// returns an empty set if one of the arguments is null
		Set<Flow> source_flows = SetUtils.getIntersection( map__server__source_flows.get( source ), outgoing_flows );
		if ( source_flows.isEmpty() ) {
			return a_out;
		} else {
			if ( source_flows != null ) {
				for ( Flow f : source_flows ) 
				{
					a_out = ArrivalCurve.add( a_out, f.getArrivalCurve() );
				}
			}			
		}
		return a_out;
	}

	// --------------------------------------------------------------------------------------------
	// paths
	// --------------------------------------------------------------------------------------------
	public Path createPath( Link link ) throws Exception {
		// Sanity check:
		if ( !links.contains( link ) ) { // Implicitly contains the case that at least one server the link connects is not in the network
			throw new Exception( "Link to create path from is not in the network" );
		}
		
		LinkedList<Server> path_servers = new LinkedList<Server>();
		path_servers.add( link.getSource() );
		path_servers.add( link.getDest() );
		
		return new Path( path_servers, new LinkedList<Link>( Collections.singleton( link ) ) );
	}
	
	public Path createPath( List<Link> path_links ) throws Exception {
		LinkedList<Server> path_servers = new LinkedList<Server>();
		
		// Sanity checks + path_servers construction:
		for( int i = 0; i < path_links.size()-1; i++ ) {
			Link l_i = path_links.get( i );
			if ( !links.contains( l_i ) ) {
				throw new Exception( "At least one link to create path from is not in the network" );
			}
			if ( l_i.getDest() != path_links.get( i+1 ).getSource() ) {
				throw new Exception( "At least two consecutive links to create the path from are not connected" );
			}
			path_servers.add( l_i.getSource() );
		}
		Link l_last = path_links.get( path_links.size()-1 );
		if ( !links.contains( l_last ) ) {
			throw new Exception( "Last link to create path from is not in the network" );
		}
		path_servers.add( l_last.getSource() );
		path_servers.add( l_last.getDest() );
		
		return new Path( path_servers, path_links );
	}

	public Path createPath( Server server ) throws Exception {
		// Sanity check:
		if ( !servers.contains( server ) ) {
			throw new Exception( "Server to create path from is not in the network");
		}
		
		return new Path( new LinkedList<Server>( Collections.singleton( server ) ), new LinkedList<Link>() );
	}
	
	public Path createPath( LinkedList<Server> path_servers ) throws Exception {
		List<Link> path_links = new LinkedList<Link>();
		
		// Sanity check + path_links construction:
		if( path_servers.size() > 1 ) {
			for ( int i = 0; i < path_servers.size()-1; i++ ) {
				try{
					path_links.add( findLink( path_servers.get( i ), path_servers.get( i+1 ) ) );
				} catch (Exception e) {
					throw new Exception( "At least two consecutive servers to create the path from are not connected by a link in the network" );
				}
			}			
		}
		
		return createPath( path_servers, path_links );
	}

	public Path createPath( LinkedList<Server> path_servers, List<Link> path_links ) throws Exception {
		// Sanity checks:
		for( int i = 0; i < path_links.size()-1; i++ ) {
			Link l_i = path_links.get( i );
			
			if ( !links.contains( l_i ) ) { // Implicitly contains the case that at least one server the link connects is not in the network
				throw new Exception( "At least one link to create path from is not in the network" );
			}
			if ( l_i.getDest() != path_links.get( i+1 ).getSource() ) {
				throw new Exception( "At least two consecutive links to create the path from are not connected" );
			}
			
			if ( path_servers.get( i ) != l_i.getSource() || path_servers.get( i+1 ) != l_i.getDest() ) {
				throw new Exception( "At least two consecutive servers to create the path from are not connected by the corresponding link of the given list" );
			}
		}
		Link l_last = path_links.get( path_links.size()-1 );
		if ( !links.contains( l_last ) ) { // Implicitly contains the case that at least one server the link connects is not in the network
			throw new Exception( "Last link to create path from is not in the network" );
		}
		
		return new Path( path_servers, path_links );
	}
	
	public Map<Server,Set<Flow>> getServerJoiningFlowsMap( Path path, Set<Flow> flows_to_join ) throws Exception {
		Map<Server,Set<Flow>> map__server__joining_flows = new HashMap<Server,Set<Flow>>();
		
		Server path_source = path.getSource();
		LinkedList<Server> servers_iteration = path.getServers();
		servers_iteration.remove( path_source );
		
		// Default for first server
		Set<Flow> flows_joining = getFlows( path_source );
		flows_joining.removeAll( flows_to_join );
		map__server__joining_flows.put( path_source, flows_joining );
		
		for ( Server s : servers_iteration ) {
			flows_joining = SetUtils.getDifference( getFlows( s ), getFlows( path.getPrecedingLink( s ) ) );
			flows_joining.removeAll( flows_to_join );
			
			// Results in an empty set if there a no joining flow at server s
			map__server__joining_flows.put( s, flows_joining );
		}

		return map__server__joining_flows;
	}

	// all flows that leave the path at the server
	public Map<Server,Set<Flow>> getServerLeavingFlowsMap( Path path ) throws Exception {
		Map<Server,Set<Flow>> map__server__leaving_flows = new HashMap<Server,Set<Flow>>();
		
		Server path_sink = path.getSink();
		LinkedList<Server> servers_iteration = path.getServers();
		servers_iteration.remove( path_sink );
		
		for ( Server s : servers_iteration ) {
			// Results in an empty set if there a no joining flow at server s
			map__server__leaving_flows.put( s, SetUtils.getDifference( getFlows( s ), getFlows( path.getSucceedingLink( s ) ) ) );
		}
		
		// Default for last server
		map__server__leaving_flows.put( path_sink, getFlows( path_sink ) );
		
		return map__server__leaving_flows;
	}
 
	/**
	 * Calculates the shortest path between src and snk
	 * according to Dijkstra's algorithm
	 * 
	 * @param src
	 * @param snk
	 * @return Dijkstra shortest path
	 * @throws Exception 
	 */
    public Path getShortestPath( Server src, Server snk ) throws Exception {
		Set<Server> visited = new HashSet<Server>();
		Map<Server,LinkedList<Link>> paths_links = new HashMap<Server,LinkedList<Link>>();
		Map<Server,LinkedList<Server>> paths_servers = new HashMap<Server,LinkedList<Server>>();

		paths_links.put( src, new LinkedList<Link>() );
		paths_servers.put( src, new LinkedList<Server>( Collections.singleton( src ) ) );
		
		LinkedList<Server> queue = new LinkedList<Server>();
		queue.add( src );
		visited.add( src );
		
		while ( !queue.isEmpty() ) {
			Server s = queue.getLast();
			queue.remove( s );
			
			Set<Server> successors_s = getSuccessors( s ); 
			for( Server successor : successors_s ) {

				LinkedList<Link> path_links_tmp = new LinkedList<Link>( paths_links.get( s ) );
				
				LinkedList<Server> path_servers_tmp;
				if ( paths_servers.containsKey( s ) ) {
					path_servers_tmp = new LinkedList<Server>( paths_servers.get( s ) );
				} else {
					path_servers_tmp = new LinkedList<Server>( Collections.singleton( src ) );
				}

				path_links_tmp.add( findLink( s, successor ) );
				path_servers_tmp.add( successor );
				
				if ( !visited.contains( successor ) ) {
					// add path
					paths_links.put( successor, path_links_tmp );
					paths_servers.put( successor, path_servers_tmp );
					
					queue.add( successor );
					visited.add( successor );
				} else {
					// compare path lengths
					if( paths_links.get( successor ).size() > path_links_tmp.size() ) {
						paths_links.put( successor, path_links_tmp );
						paths_servers.put( successor, path_servers_tmp );
						
						// TODO is it necessary to re-check the path to the successor?
						queue.add( successor );
					}
				}				
			}
		}
		
		if ( paths_links.get( snk ) == null ) {
			throw new Exception( "No path from server " + src.getId() + " to server " + snk.getId() + " found" );
		}
		
		// No sanity checks needed after a shortest path calculation, so you can create a new path directly instead of calling 'createPath'
		return new Path( paths_servers.get( snk ), paths_links.get( snk ) );
	}
 	
 	/**
 	 * Returns the server at which the flows in <code>flows_of_interest</code>
 	 * first all meet each other (when viewed from the source). When viewed from
 	 * <code>server</code> towards the sink, this is the last server where all
 	 * flows are still together.
 	 * 
 	 * @param server_common_dest
 	 * @param flows_of_interest
 	 * @return the splitting point server
 	 */
 	public Server findSplittingServer( Server server_common_dest, Set<Flow> flows_of_interest ) throws Exception {
 		Flow f = flows_of_interest.iterator().next();
 		
 		if ( flows_of_interest.size() == 1 ) {
 			return f.getSource();
 		}
 		
 		Path f_path = f.getPath();
 		int common_dest_index_f = f_path.getServers().indexOf( server_common_dest );
 		
 		Server split = server_common_dest;
 		// Iterate in reverse order starting from server_common_dest, stop as soon as at least one of the flows of interest is missing
 		for( int i = common_dest_index_f - 1; i >= 0; i-- ) { // -1 excludes server_common_dest
 			Server split_candidate = f_path.getServers().get( i );
 			
 			if ( getFlows( split_candidate ).containsAll( flows_of_interest ) ) {
 				split = split_candidate;
 			} else {
 				break;
 			}
 		}
 		
 		if ( split == server_common_dest ) { // no split found
 	 		throw new Exception( "Could not find splitting point" );
 		} else {
 			return split;
 		}
 	}

	// --------------------------------------------------------------------------------------------
	// other helper functions
	// --------------------------------------------------------------------------------------------
 	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given lnik path <code>path</code> that have the useGamma flag set.<br />
	 * If a server either has no maximum service curve set or useGamma is disabled,
	 * calculations take place with the default maximum service curve,
	 * i.e., the zero delay burst curve, so the result will not be influenced.
	 * 
	 * @param path
	 *            an link path
	 * @return the convolved curve
	 */
	public static ServiceCurve getGamma( Path path )
	{
		Collection<Server> servers = path.getServers();
		return getGamma( servers );
	}
	
	public static ServiceCurve getGamma( Collection<Server> servers )
	{
		ServiceCurve gamma_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			gamma_total = Convolution.convolve( gamma_total, s.getGamma() );
		}
		
		return gamma_total;
	}

	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given link path <code>path</code> that have the useExtraGamma flag set.<br />
	 * If a server either has no maximum service curve set or useExtraGamma is disabled,
	 * calculations take place with the default maximum service curve,
	 * i.e., the zero delay burst curve, so the result will not be influenced.
	 * 
	 * @param path
	 *            an link path
	 * @return the convolved curve
	 */
	public static ServiceCurve getExtraGamma( Path path )
	{
		Collection<Server> servers = path.getServers();
		return getExtraGamma( servers );
	}

	public static ServiceCurve getExtraGamma( Collection<Server> servers )
	{
		ServiceCurve extra_gamma_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			extra_gamma_total = Convolution.convolve( extra_gamma_total, s.getExtraGamma() );
		}	
		extra_gamma_total.removeLatency();
		
		return extra_gamma_total;
	}
	
	/**
	 * Returns the convolution of the maximum service curves of all servers on
	 * the given link path <code>path</code><br />
	 * If a server has no maximum service curve, calculations take place with 
	 * the default maximum service curve, i.e., the zero delay burst curve, 
	 * so the result will not be influenced.
	 * 
	 * @param path
	 *            an link path
	 * @return the convolved curve
	 */
	public static ServiceCurve getMaxServiceCurve( Path path )
	{
		Collection<Server> servers = path.getServers();
		return getMaxServiceCurve( servers );
	}

	public static ServiceCurve getMaxServiceCurve( Collection<Server> servers )
	{
		ServiceCurve max_service_curve_total = ServiceCurve.createZeroDelayBurst();
		for ( Server s : servers )
		{
			max_service_curve_total = Convolution.convolve( max_service_curve_total, s.getMaxServiceCurve() );
		}
		
		return max_service_curve_total;
	}
}
